"""example URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.2/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.conf import settings
from django.conf.urls.static import static
from django.contrib import admin
from django.urls import include, path

OIDC_URL_PREFIX = getattr(settings, 'OIDC_URL_PREFIX', '')

urlpatterns = [
    path('admin/', admin.site.urls),
]

urlpatterns += static(settings.STATIC_URL,
                      document_root=settings.STATIC_ROOT)

if 'oidc_provider' in settings.INSTALLED_APPS:
    import oidc_provider.urls
    from oidc_provider.views import well_known
    urlpatterns += path('.well-known/<str:service>',
                        well_known, name="oidc_op_well_known"),
    urlpatterns += path(f'{OIDC_URL_PREFIX}',
                        include((oidc_provider.urls, 'oidc_op',))),



import base64
import logging
import json
import os
import sys
import urllib

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.http import (HttpResponse,
                         HttpResponseBadRequest,
                         HttpResponseForbidden,
                         HttpResponseRedirect,
                         JsonResponse)
from django.views.decorators.csrf import csrf_exempt
from django.shortcuts import render
from oidcop.authn_event import create_authn_event
from oidcop.exception import InvalidClient
from oidcop.exception import UnAuthorizedClient
from oidcop.exception import UnknownClient
from oidcmsg.oidc import AuthorizationErrorResponse
from oidcop.oidc.token import Token
from oidcmsg.oauth2 import ResponseMessage
from oidcmsg.oidc import AccessTokenRequest
from oidcmsg.oidc import AuthorizationRequest
from urllib.parse import urlparse

from . application import oidcop_app
from . decorators import fill_cdb_by_request, prepare_oidc_endpoint, debug_request
from . exceptions import InconsinstentSessionDump
from . models import OidcRelyingParty, OidcSession, OidcIssuedToken


logger = logging.getLogger(__name__)

IGNORED_HEADERS = ["cookie", "user-agent"]


def _add_cookie(resp, cookie_spec):
    kwargs = {
        k: v
        for k, v in cookie_spec.items()
        if k not in ('name',)
    }
    kwargs["path"] = "/"
    resp.set_cookie(cookie_spec["name"], **kwargs)


def add_cookie(resp, cookie_spec):
    if isinstance(cookie_spec, list):
        for _spec in cookie_spec:
            _add_cookie(resp, _spec)
    elif isinstance(cookie_spec, dict):
        _add_cookie(resp, cookie_spec)


def _check_session_dump_consistency(endpoint_name, ec, session):
    if ec.endpoint_context.session_manager.dump() != session:
        logger.critical(
            ec.endpoint_context.session_manager.dump(),
            session
        )
        ec.endpoint_context.session_manager.flush()
        raise InconsinstentSessionDump(endpoint_name)


def do_response(request, endpoint, req_args, error='', **args):
    info = endpoint.do_response(request=req_args, error=error, **args)
    _response_placement = info.get('response_placement')
    if not _response_placement:
        _response_placement = endpoint.response_placement

    info_response = info['response']

    if settings.DEBUG:
        # Debugging things
        try:
            response_params = json.dumps(json.loads(info_response), indent=2)
            logger.debug('Response params: {}\n'.format(response_params))
        except Exception:
            url, args = urllib.parse.splitquery(info_response)
            response_params = urllib.parse.parse_qs(args)
            resp = json.dumps(response_params, indent=2)
            logger.debug('Response params: {}\n{}\n\n'.format(url, resp))
        # end debugging

    if error:
        if _response_placement == 'body':
            logger.error('Error Response [Body]: {}'.format(info_response))
            resp = HttpResponse(info_response, status=400)
        else:  # _response_placement == 'url':
            logger.debug('Redirect to: {}'.format(info_response))
            resp = HttpResponseRedirect(info_response)
    else:
        if _response_placement == 'body':
            # logger.debug('Response [Body]: {}'.format(info_response))
            resp = HttpResponse(info_response, status=200)
        else:  # _response_placement == 'url':
            # logger.debug('Redirect to: {}'.format(info_response))
            resp = HttpResponseRedirect(info_response)

    for key, value in info['http_headers']:
        # set response headers
        resp[key] = value

    if 'cookie' in info:
        add_cookie(resp, info['cookie'])

    ec = oidcop_app.endpoint_context
    ses_man_dump = ec.endpoint_context.session_manager.dump()

    # session db mngmtn
    if endpoint.__class__.__name__ in ('Authorization',
                                       'Token',
                                       'UserInfo'):
        try:
            OidcSession.load(ses_man_dump)
        except InconsinstentSessionDump as e:
            logger.critical(e)
            ec.endpoint_context.session_manager.flush()
            return JsonResponse(json.dumps({
                'error': 'invalid_request',
                'error_description': str(e),
                'method': request.method
            }), safe=False, status=500)
        else:
            #  logger.warning(endpoint.__class__.__name__)
            _check_session_dump_consistency(endpoint.__class__.__name__,
                                            ec,
                                            ses_man_dump)
        ec.endpoint_context.session_manager.flush()
    return resp


def _get_http_info(request):
    http_info = {
        "headers": {
            k.lower(): v
            for k, v in request.headers.items()
            if k not in IGNORED_HEADERS
        },
        "method": request.method,
        "url": request.get_raw_uri(),
        # name is not unique
        "cookie": [
            {"name": k, "value": v} for k, v in request.COOKIES.items()
        ]
    }
    return http_info


def _get_http_data(request, http_info):
    data = {}
    if request.method == 'GET':
        data = {k: v for k, v in request.GET.items()}
    elif request.method == 'POST':
        data = {k: v for k, v in request.POST.items()}

    if not data and request.body:
        data = json.loads(request.body)

    return data


def service_endpoint(request, endpoint):
    """
    TODO: documentation here
    """
    logger.info('Request at the "{}" endpoint'.format(endpoint.name))

    http_info = _get_http_info(request)
    data = _get_http_data(request, http_info)
    req_args = endpoint.parse_request(data, http_info=http_info)

    try:
        if isinstance(endpoint, Token):
            args = endpoint.process_request(
                AccessTokenRequest(**req_args), http_info=http_info
            )
        else:
            args = endpoint.process_request(req_args, http_info=http_info)
    except (InvalidClient, UnknownClient, UnAuthorizedClient) as err:
        logger.error(err)
        return JsonResponse(json.dumps({
            'error': 'unauthorized_client',
            'error_description': str(err)
        }), safe=False, status=400)
    except Exception as err:
        logger.error(err)
        return JsonResponse(json.dumps({
            'error': 'invalid_request',
            'error_description': str(err),
            'method': request.method
        }), safe=False, status=400)

    if isinstance(req_args, ResponseMessage) and 'error' in req_args:
        return JsonResponse(req_args.__dict__, status=400)
    elif 'redirect_location' in args:
        return HttpResponseRedirect(args['redirect_location'])
    elif 'http_response' in args:
        return HttpResponse(args['http_response'], status=200)

    return do_response(request, endpoint, req_args, **args)


def well_known(request, service):
    """
    /.well-known/<service>
    """
    _name = sys._getframe().f_code.co_name
    debug_request(f'{_name}', request)

    if service == 'openid-configuration':
        _endpoint = oidcop_app.endpoint_context.endpoint['provider_config']
    # TODO fedservice integration here
    # if service == 'openid-federation':
    #     _endpoint = oidcop_app.endpoint_context.endpoint['provider_info']
    elif service == 'webfinger':
        _endpoint = oidcop_app.endpoint_context.endpoint['webfinger']
    else:
        return HttpResponseBadRequest('Not supported', status=400)

    return service_endpoint(request, _endpoint)


@csrf_exempt
@prepare_oidc_endpoint
def registration(request):
    ec = oidcop_app.endpoint_context
    _endpoint = ec.endpoint['registration']
    response = service_endpoint(request, _endpoint)
    # update db
    OidcRelyingParty.import_from_cdb(
        oidcop_app.endpoint_context.endpoint_context.cdb
    )
    return response


@csrf_exempt
@prepare_oidc_endpoint
@fill_cdb_by_request
def registration_read(request):
    return service_endpoint(
        request,
        oidcop_app.endpoint_context.endpoint['registration_read']
    )


def _fill_cdb_by_client(client):
    ec = oidcop_app.endpoint_context
    ec.endpoint_context.cdb = {
        client.client_id: client.serialize()
    }


@prepare_oidc_endpoint
@fill_cdb_by_request
def authorization(request):
    ec = oidcop_app.endpoint_context
    _endpoint = ec.endpoint['authorization']
    return service_endpoint(request, _endpoint)


@csrf_exempt
def verify_user(request):
    """csrf is not needed because it uses oidc token in the post
    """
    _name = sys._getframe().f_code.co_name
    debug_request(f'{_name}', request)

    token = request.POST.get('token')
    if not token:  # pragma: no cover
        return HttpResponse('Access forbidden: invalid token.', status=403)

    ec = oidcop_app.endpoint_context
    authn_method = ec.endpoint_context.authn_broker.get_method_by_id('user')

    kwargs = dict([(k, v) for k, v in request.POST.items()])
    user = authn_method.verify(**kwargs)
    if not user:
        return HttpResponse('Authentication failed', status=403)

    auth_args = authn_method.unpack_token(kwargs['token'])
    authz_request = AuthorizationRequest().from_urlencoded(auth_args['query'])

    # salt size can be customized in settings.OIDC_OP_AUTHN_SALT_SIZE
    salt_size = getattr(settings, 'OIDC_OP_AUTHN_SALT_SIZE', 4)
    authn_event = create_authn_event(
        uid=user.username,
        salt=base64.b64encode(os.urandom(salt_size)).decode(),
        authn_info=auth_args['authn_class_ref'],
        authn_time=auth_args['iat']
    )

    endpoint = oidcop_app.endpoint_context.endpoint['authorization']
    client_id = authz_request["client_id"]
    _token_usage_rules = endpoint.server_get(
        "endpoint_context").authn_broker.get_method_by_id('user')

    session_manager = ec.endpoint_context.session_manager
    _session_id = session_manager.create_session(
        authn_event=authn_event,
        auth_req=authz_request,
        user_id=user.username,
        client_id=client_id,
        token_usage_rules=_token_usage_rules
    )

    try:
        _args = endpoint.authz_part2(user=user.username,
                                     session_id=_session_id,
                                     request=authz_request,
                                     authn_event=authn_event)
    except ValueError as excp:
        msg = 'Something went wrong with your Session ... {}'.format(excp)
        return HttpResponse(msg, status=403)

    if isinstance(_args, ResponseMessage) and 'error' in _args:
        return HttpResponse(_args.to_json(), status=400)
    elif isinstance(_args.get('response_args'), AuthorizationErrorResponse):
        rargs = _args.get('response_args')
        logger.error(rargs)
        return HttpResponse(rargs.to_json(), status=400)

    response = do_response(request, endpoint, authz_request, **_args)
    return response


def _get_session_by_token(request):
    bearer = request.META.get('HTTP_AUTHORIZATION')
    token = None

    if bearer and not request.POST:
        token = OidcIssuedToken.objects.filter(
            value=bearer.split(' ')[1]
        ).first()
    elif all((request.POST.get('refresh_token'),
              request.POST.get('grant_type') == 'refresh_token')):
        token = OidcIssuedToken.objects.filter(
            type=request.POST['grant_type'],
            value=request.POST['refresh_token']
        ).first()
    elif request.POST.get('grant_type') == 'authorization_code':
        token = OidcIssuedToken.objects.filter(
            type=request.POST['grant_type'],
            value=request.POST.get('code')
        ).first()
    elif request.POST.get('token_type_hint'):
        # token introspection
        token = OidcIssuedToken.objects.filter(
            type=request.POST['token_type_hint'],
            value=request.POST['token']
        ).first()
    elif request.GET.get('id_token_hint') or request.POST.get('id_token_hint'):
        _id_token_hint = request.GET.get('id_token_hint') or request.POST.get('id_token_hint')
        token = OidcIssuedToken.objects.filter(
            type='id_token',
            value=_id_token_hint
        ).first()
    else:
        raise PermissionDenied()

    if token:
        return token.session
    else:
        raise PermissionDenied()


@csrf_exempt
@prepare_oidc_endpoint
@fill_cdb_by_request
def token(request):
    ec = oidcop_app.endpoint_context
    _endpoint = ec.endpoint['token']
    session = _get_session_by_token(request).serialize()
    if session:
        ec.endpoint_context.session_manager.load(session)
    _check_session_dump_consistency('token', ec, session)

    response = service_endpoint(request, _endpoint)
    return response


@csrf_exempt
@prepare_oidc_endpoint
def userinfo(request):
    ec = oidcop_app.endpoint_context
    _endpoint = ec.endpoint['userinfo']

    session = _get_session_by_token(request)
    _fill_cdb_by_client(session.client)
    session = session.serialize()

    if session:
        ec.endpoint_context.session_manager.load(session)
    _check_session_dump_consistency('userinfo', ec, session)

    return service_endpoint(request, _endpoint)


@csrf_exempt
@prepare_oidc_endpoint
def introspection(request):
    ec = oidcop_app.endpoint_context
    _endpoint = ec.endpoint['introspection']

    session = _get_session_by_token(request)
    _fill_cdb_by_client(session.client)
    ec.endpoint_context.cdb
    session = session.serialize()

    if session:
        ec.endpoint_context.session_manager.load(session)
    _check_session_dump_consistency('introspection', ec, session)

    return service_endpoint(request, _endpoint)


########
# LOGOUT
########
@prepare_oidc_endpoint
def session_endpoint(request):
    ec = oidcop_app.endpoint_context
    _endpoint = ec.endpoint['session']
    session = _get_session_by_token(request)
    _fill_cdb_by_client(session.client)
    session = session.serialize()
    if session:
        ec.endpoint_context.session_manager.load(session)
    _check_session_dump_consistency('session', ec, session)
    try:
        res = service_endpoint(request, _endpoint)
        return res
    except Exception:
        ec.endpoint_context.session_manager.flush()
        return HttpResponseForbidden()


# TODO - not supported yet with session manager storage

def check_session_iframe(request):
    if request.method == 'GET':
        req_args = request.GET
    elif request.method == 'POST':
        req_args = json.loads(request.POST)
    else:
        req_args = dict([(k, v) for k, v in request.body.items()])

    if req_args:
        # will contain client_id and origin
        if req_args['origin'] != oidcop_app.endpoint_context.conf['issuer']:
            return 'error'
        if req_args['client_id'] != oidcop_app.endpoint_context.cdb:
            return 'error'
        return 'OK'

    logger.debug('check_session_iframe: {}'.format(req_args))
    res = render(request, template_name='check_session_iframe.html')
    return res


@csrf_exempt
def rp_logout(request):
    _name = sys._getframe().f_code.co_name
    debug_request(f'{_name}', request)

    _endp = oidcop_app.endpoint_context.endpoint['session']
    _info = _endp.unpack_signed_jwt(request.POST['sjwt'])
    alla = None

    _iframes = _endp.do_verified_logout(alla=alla, **_info)
    if _iframes:
        d = dict(frames=" ".join(_iframes),
                 size=len(_iframes),
                 timeout=5000,
                 postLogoutRedirectUri=_info['redirect_uri'])
        res = render(request, 'frontchannel_logout.html', d)

    else:
        res = HttpResponseRedirect(_info['redirect_uri'])
        try:
            _endp.kill_cookies()
        except AttributeError:
            logger.debug('Cookie not implemented or not working.')
    return res


def verify_logout(request):
    _name = sys._getframe().f_code.co_name
    debug_request(f'{_name}', request)

    part = urlparse(oidcop_app.endpoint_context.conf['issuer'])
    d = dict(op=part.hostname,
             do_logout='rp_logout',
             sjwt=request.GET['sjwt'] or request.POST['sjwt'])
    return render(request, 'logout.html', d)


def post_logout(request):
    _name = sys._getframe().f_code.co_name
    debug_request(f'{_name}', request)
    return render(request, 'post_logout.html')



# Generated by Django 3.1.1 on 2021-05-15 10:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('oidc_provider', '0005_oidcsession_sid_encrypted'),
    ]

    operations = [
        migrations.AlterField(
            model_name='oidcrpgranttype',
            name='grant_type',
            field=models.CharField(choices=[('authorization_code', 'authorization_code'), ('urn:ietf:params:oauth:grant-type:jwt-bearer',
                                                                                           'urn:ietf:params:oauth:grant-type:jwt-bearer'), ('refresh_token', 'refresh_token')], max_length=60),
        ),
    ]



# Generated by Django 3.1.1 on 2021-05-10 13:53

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('oidc_provider', '0002_oidcsession_salt'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='oidcissuedtoken',
            options={'verbose_name': 'Issued Token',
                     'verbose_name_plural': 'Issued Tokens'},
        ),
    ]



import json
import logging

from django import forms
from django.contrib import admin
from django.contrib.sessions.models import Session
from django.utils.safestring import mark_safe

from . models import OidcIssuedToken
from . models import OidcRPContact
from . models import OidcRPRedirectUri
from . models import OidcRPGrantType
from . models import OidcRPResponseType
from . models import OidcRPScope
from . models import OidcRelyingParty
from . models import OidcSession

logger = logging.getLogger(__name__)


class OidcRPContactModelForm(forms.ModelForm):
    class Meta:
        model = OidcRPContact
        fields = ('__all__')


class OidcRPContactInline(admin.TabularInline):
    model = OidcRPContact
    form = OidcRPContactModelForm
    extra = 0


class OidcRPRedirectUriModelForm(forms.ModelForm):
    class Meta:
        model = OidcRPRedirectUri
        fields = ('__all__')


class OidcRPRedirectUriInline(admin.TabularInline):
    model = OidcRPRedirectUri
    form = OidcRPRedirectUriModelForm
    extra = 0


class OidcRPGrantTypeModelForm(forms.ModelForm):
    class Meta:
        model = OidcRPGrantType
        fields = ('__all__')


class OidcRPGrantTypeInline(admin.TabularInline):
    model = OidcRPGrantType
    form = OidcRPGrantTypeModelForm
    extra = 0


class OidcRPResponseTypeModelForm(forms.ModelForm):
    class Meta:
        model = OidcRPResponseType
        fields = ('__all__')


class OidcRPResponseTypeInline(admin.TabularInline):
    model = OidcRPResponseType
    form = OidcRPResponseTypeModelForm
    extra = 0


class OidcRPScopeModelForm(forms.ModelForm):
    class Meta:
        model = OidcRPScope
        fields = ('__all__')


class OidcRPScopeInline(admin.TabularInline):
    model = OidcRPScope
    form = OidcRPScopeModelForm
    extra = 0


@admin.register(OidcRelyingParty)
class OidcRelyingPartyAdmin(admin.ModelAdmin):
    list_filter = ('created', 'modified', 'is_active',
                   'client_id_issued_at', 'client_secret_expires_at')
    list_display = ('client_id', 'created',
                    'last_seen', 'is_active')
    search_fields = ('client_id',)
    list_editable = ('is_active',)
    inlines = (OidcRPScopeInline,
               OidcRPResponseTypeInline,
               OidcRPGrantTypeInline,
               OidcRPContactInline,
               OidcRPRedirectUriInline)
    fieldsets = (
        (None, {
            'fields': (
                ('client_id', 'client_secret',),
                ('client_salt', 'jwks_uri'),
                ('registration_client_uri',),
                ('registration_access_token',),
                ('application_type',
                 'token_endpoint_auth_method'),
                ('is_active', )
            )
        },
        ),
        ('Temporal values',
         {
             'fields': (
                 (('client_id_issued_at',
                   'client_secret_expires_at',
                   'last_seen')),

             ),

         },
         ),
    )


@admin.register(Session)
class SessionAdmin(admin.ModelAdmin):  # pragma: no cover
    def _session_data(self, obj):
        return obj.get_decoded()
    list_display = ['session_key', '_session_data', 'expire_date']


@admin.register(OidcIssuedToken)
class OidcIssuedTokenAdmin(admin.ModelAdmin):
    search_fields = ('value', 'session__user__username')
    list_display = ['session', 'type', 'created']
    list_filter = ('issued_at', 'expires_at', 'revoked', 'type')
    readonly_fields = (
        "type",
        "issued_at",
        "expires_at",
        "not_before",
        "revoked",
        "value",
        "usage_rules",
        "used",
        "based_on",
        "session",
        "uid"
    )


@admin.register(OidcSession)
class OidcSessionAdmin(admin.ModelAdmin):
    list_filter = ('created', 'modified', 'expires_at')
    list_display = ('user', 'user_uid', 'client',
                    'grant_uid', 'created', 'expires_at')
    search_fields = ('user__username', 'client__client_id')
    readonly_fields = ('user_uid', 'user', 'client',
                       'sub', 'sid', 'sid_encrypted', 'key', 'salt',
                       'created', 'expires_at',
                       'user_session_info_preview',
                       'client_session_info_preview',
                       'grant_preview'
                       )

    fieldsets = (
        (None, {
            'fields': (
                ('client', ),
                ('user_uid',),
                ('sub', ),
                ('sid',),
                ('sid_encrypted',),
                ('key',),
                ('salt',),
                ('created',),
                ('expires_at',),
            )
        },
        ),

        ('User session info',
         {
             # 'classes': ('collapse',),
             'fields': (
                 ('user_session_info_preview'),
             )

         },
         ),

        ('Client session info',
         {
             # 'classes': ('collapse',),
             'fields': (
                 ('client_session_info_preview'),
             )

         },
         ),

        ('Grant session info',
         {
             # 'classes': ('collapse',),
             'fields': (
                 ('grant_preview'),
             )
         },
         ),
    )

    def user_session_info_preview(self, obj):
        dumps = json.dumps(obj.user_session_info, indent=2)
        return mark_safe(dumps.replace('\n', '<br>').replace(r' ', '&nbsp'))
    user_session_info_preview.short_description = 'User Session Info'

    def client_session_info_preview(self, obj):
        dumps = json.dumps(obj.client_session_info, indent=2)
        return mark_safe(dumps.replace('\n', '<br>').replace(r' ', '&nbsp'))
    client_session_info_preview.short_description = 'Client Session Info'

    def grant_preview(self, obj):
        dumps = json.dumps(obj.grant, indent=2)
        return mark_safe(dumps.replace('\n', '<br>').replace(r' ', '&nbsp'))
    grant_preview.short_description = 'Grant'

    class Media:
        js = ('js/textarea_autosize.js',)



import base64
import datetime
import logging
import re
import urllib


from django.contrib.auth import get_user_model
from django.test import Client
from django.test import TestCase
from django.urls import reverse
from django.utils import timezone

from oidc_provider.models import OidcRelyingParty
from oidc_provider.utils import decode_token, dt2timestamp, timestamp2dt

logger = logging.getLogger('oidc_provider')

CLIENT_1_ID = 'jbxedfmfyc'
CLIENT_1_PASSWD = '19cc69b70d0108f630e52f72f7a3bd37ba4e11678ad1a7434e9818e1'
CLIENT_1_RAT = 'z3PCMmC1HZ1QmXeXGOQMJpWQNQynM4xY'
CLIENT_1_SESLOGOUT = 'https://127.0.0.1:8099/session_logout/django_provider'
CLIENT_1 = {
    'jbxedfmfyc': {
        'client_id': CLIENT_1_ID,
        'client_salt': '6flfsj0Z',
        'registration_access_token': CLIENT_1_RAT,
        'registration_client_uri': 'https://127.0.0.1:8000/registration_api?client_id=jbxedfmfyc',
        'client_id_issued_at': timezone.localtime().timestamp(),
        'client_secret': CLIENT_1_PASSWD,
        'client_secret_expires_at': (timezone.localtime() + timezone.timedelta(days=1)).timestamp(),
        'application_type': 'web',
        'contacts': ['ops@example.com'],
        'token_endpoint_auth_method': 'client_secret_basic',
        # 'jwks_uri': 'https://127.0.0.1:8099/static/jwks.json',
        'redirect_uris': [('https://127.0.0.1:8099/authz_cb/django_provider', {})],
        'post_logout_redirect_uris': [(CLIENT_1_SESLOGOUT, None)],
        'response_types': ['code'],
        'grant_types': ['authorization_code'],
        'allowed_scopes': ['openid', 'profile', 'email', 'offline_access']
    }
}
CLIENT_1_BASICAUTHZ = f'Basic {base64.b64encode(f"{CLIENT_1_ID}:{CLIENT_1_PASSWD}".encode()).decode()}'
ACCESS_TOKEN = "eyJhbGciOiJFUzI1NiIsImtpZCI6Ik16UldNamhvUmt0UVVYWklkRTFSZUV0cE5GRkJTMFIwVEc1T05WVlJUazltVkZOWFFqVm9VMHBQWncifQ.eyJzY29wZSI6IFsib3BlbmlkIiwgInByb2ZpbGUiLCAiZW1haWwiLCAiYWRkcmVzcyIsICJwaG9uZSIsICJvZmZsaW5lX2FjY2VzcyJdLCAiYXVkIjogWyI2TXdoRzNWRl9BWml5U2huSHlteWxRIl0sICJqdGkiOiAiYmY5N2I0MThjOTZlMTFlYmEwYjAzZjlmOWRhN2RhNjkiLCAiY2xpZW50X2lkIjogIjZNd2hHM1ZGX0FaaXlTaG5IeW15bFEiLCAic3ViIjogIjZNd2hHM1ZGX0FaaXlTaG5IeW15bFEiLCAic2lkIjogIlowRkJRVUZCUW1kM1ZITlRibmx2Y0dOdFNrbFVRazFvZDBSWlUzcGFSR2hmT1hkWFNtaElWelZhVW05NmFHWkxhRVU1Y2xSc1FXVkJSMjlWVVZCWFZXVjFOazlCVERrNVN6TkZWMWRZTTBkaVVtbFZMVVV5ZVdKMVdXOUZPVWxKVDBoZk4zcDFkWEZMZDI1MGMwUm1Ua0YzUWpWVk4yNXdkekZYTUU1MkxWZFhjMkp5Y1VwQ2RFZG5jRTFXYUc1eGFIUjNkVGRCVkcxWFQyNURSR1kwZFZScU0yOWlZMFpYTTJnNFRVNWFObmRDVGxabFdFbEJRVkpVZHpKeldrOHhWWHBTWWtKMVdFY3phRXBQVTBKU1VXSmtaaTFVV1ZrMGEwVTBhWFpDZGxONFJ6UnFaWE5RWTBGTVRqaDZOek56YWpkTGFsaFBZejA9IiwgInRva2VuX2NsYXNzIjogImFjY2Vzc190b2tlbiIsICJpc3MiOiAiaHR0cHM6Ly8xMjcuMC4wLjE6ODAwMCIsICJpYXQiOiAxNjIzMjc2MzA2LCAiZXhwIjogMTYyMzI3OTkwNn0.tbJVJVKIgqhkD8cDldteeQY3FyLgckvggD-dWqjHbHB2HRo0lqv17DHGOOs4O5HwY3YMLG_yibgz8ncaSj6MYw"


class TestOidcRPFlow(TestCase):
    def setUp(self):
        self.client = Client()

    def _test_discovery_provider(self):
        url = reverse('oidc_provider:_well_known',
                      kwargs={'service': 'openid-configuration'}
                      )
        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(type(response.json()), dict)

    def test_registration(self):
        url = reverse('oidc_provider:registration')
        data = {
            "application_type": "web",
            "response_types": ["code"],
            "contacts": ["ops@example.com"],
            "token_endpoint_auth_method": "client_secret_basic",
            "redirect_uris": ["https://127.0.0.1:8099/authz_cb/django_provider"],
            "post_logout_redirect_uris": ["https://127.0.0.1:8099/session_logout/django_provider"],
            # "jwks_uri": "https://127.0.0.1:8099/static/jwks.json",
            "frontchannel_logout_uri": "https://127.0.0.1:8099/fc_logout/django_provider",
            "backchannel_logout_uri": "https://127.0.0.1:8099/bc_logout/django_provider",
            "grant_types": ["authorization_code"]
        }
        response = self.client.post(url, data=data)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(type(response.json()), dict)

        # test read registration api
        OidcRelyingParty.import_from_cdb(CLIENT_1)
        url = reverse('oidc_provider:registration_read')
        headers = {
            'HTTP_AUTHORIZATION': f"Bearer {response.json()['registration_access_token']}"
        }
        response = self.client.get(
            url,
            {'client_id': response.json()['client_id']},
            **headers
        )
        self.assertEqual(response.status_code, 200)
        self.assertIn('client_secret', response.json())

    def test_authz(self):
        # without dynamic client registration here ...
        OidcRelyingParty.import_from_cdb(CLIENT_1)
        data = {
            'redirect_uri': 'https://127.0.0.1:8099/authz_cb/django_provider',
            'scope': 'openid profile email address phone offline_access',
            'response_type': 'code',
            'nonce': 'd0GjnvG2zVZmnbyBzioXOFXU',
            'state': 'AREIA28Qh8xICz4j3rSc2tSQkV3QbFvx',
            'code_challenge': 'Mbp5kPfcaQ6MVj8MUcjcjHEcbwuNY5GBrG5HqJLgjCg',
            'code_challenge_method': 'S256',
            'client_id': 'jbxedfmfyc',
            'prompt': 'consent'
        }
        url = reverse('oidc_provider:authorization')
        response = self.client.get(url, data=data)
        print(f'Authorization request: {response.request}')
        self.assertEqual(response.status_code, 200)
        self.assertIn('form action="/oidcop/verify/oidc_user_login/',
                      response.content.decode())

        # yes I do a form submission now
        auth_code = re.search(
            'value="(?P<token>[a-zA-Z\-\.\_0-9]*)"',
            response.content.decode()).groupdict()['token']

        auth_dict = {
            'username': 'test',
            'password': 'testami18_WRONG',
            'token': auth_code
        }

        user = get_user_model().objects.create(
            username='test',
            email='me@my.self',
            is_staff=1,
            is_superuser=1)
        user.set_password('testami18')
        user.save()

        url = reverse('oidc_provider:verify_user')
        response = self.client.post(url, data=auth_dict)
        self.assertEqual(response.status_code, 403)

        auth_dict['password'] = 'testami18'
        response = self.client.post(url, data=auth_dict)

        # put the right cookie in
        _cookies = response.cookies

        self.assertEqual(response.status_code, 302)
        print(f'Authorization code redirect: {response.url}')

        _data = urllib.parse.parse_qs(response.url.split('?')[1])

        data = {
            'grant_type': 'authorization_code',
            'redirect_uri': data['redirect_uri'],
            'client_id': CLIENT_1_ID,
            'state': _data['state'],
            'code': _data['code'],
            'code_verifier': '8TuTiNE18VhP2B3wtxfz8pjr1Y8bAxNjJ3VE7GjVHVWPCZ6mdYEnkbaYRquw9Fna'
        }

        url = reverse('oidc_provider:token')
        headers = {
            'HTTP_AUTHORIZATION': CLIENT_1_BASICAUTHZ
        }
        response = self.client.post(url, data=data, **headers)

        self.assertEqual(response.status_code, 200)
        self.assertIn('refresh_token', response.json())
        self.assertIn('access_token', response.json())

        _id_token = response.json()['id_token']

        # refresh token
        data = {
            "grant_type" : "refresh_token",
            "client_id" : f"{CLIENT_1_ID}",
            "client_secret" : f"{CLIENT_1_PASSWD}",
            "refresh_token" : f"{response.json()['refresh_token']}"
        }
        response = self.client.post(url, data=data, **headers)
        self.assertIn('access_token', response.json())

        # userinfo
        url = reverse('oidc_provider:userinfo')
        headers = {
            'HTTP_AUTHORIZATION': f"Bearer {response.json()['access_token']}"
        }
        response = self.client.get(url, **headers)

        # test admin
        self.client.login(username='test', password='testami18')
        url = reverse('admin:oidc_provider_oidcsession_change',
                      kwargs={'object_id': 1})
        response = self.client.get(url)
        # end admin

        # session logout
        data = {
            'id_token_hint': _id_token,
            'post_logout_redirect_uri': 'https://127.0.0.1:8099/session_logout/django_provider',
            'state': _data['state'][0]
        }
        headers = {
            'HTTP_COOKIE': f"oidc_op='{_cookies.get('oidc_op').value}'"
        }
        url = reverse('oidc_provider:session')
        response = self.client.get(
            url, data=data, cookies=_cookies, **headers
        )
        self.assertEqual(response.status_code, 302)

        # verify logout
        url = reverse('oidc_provider:verify_logout')
        _qs = response.url.split('?')[1]
        response = self.client.get(f"{url}?{_qs}")
        self.assertIn('type="hidden" name="sjwt"', response.content.decode())

        # rp logout
        url = reverse('oidc_provider:rp_logout')
        data = {
            'sjwt': [urllib.parse.parse_qs(_qs)['sjwt'][0]],
            'logout': ['yes']
        }
        response = self.client.post(url, data=data, **headers)
        self.assertEqual(response.status_code, 302)
        self.assertIn(CLIENT_1_SESLOGOUT, response.url)
        self.assertIn(_data['state'][0], response.url)


    def test_utils(self):
        decode_token(ACCESS_TOKEN)

        now = datetime.datetime.now()
        ts = dt2timestamp(now)
        timestamp2dt(ts)



# Generated by Django 3.1.1 on 2021-05-16 12:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('oidc_provider', '0006_auto_20210515_1001'),
    ]

    operations = [
        migrations.AlterField(
            model_name='oidcissuedtoken',
            name='value',
            field=models.TextField(unique=True),
        ),
    ]



from django.contrib import admin
from django.utils.translation import ugettext_lazy as _
from django.contrib.auth.admin import UserAdmin

from .models import User, PersistentId
from .admin_inlines import PersistentIdInline


@admin.register(User)
class CustomUserAdmin(UserAdmin):
    readonly_fields = ('date_joined', 'last_login',)
    list_display = ('username', 'email', 'is_active',
                    'is_staff', 'is_superuser', )
    list_editable = ('is_active', 'is_staff', 'is_superuser',)
    inlines = [PersistentIdInline, ]
    fieldsets = (
        (None, {'fields': (('username', 'is_active', 'is_staff', 'is_superuser', ),
                           ('password'),
                           ('origin'),
                           )
                }),
        (_('Anagrafica'), {'fields': (('first_name', 'last_name'),
                                      'email',
                                      ('taxpayer_id',),
                                      ('gender',
                                          'place_of_birth', 'birth_date',),
                                      )
                           }),

        (_('Permissions'), {'fields': ('groups', 'user_permissions'),
                            'classes': ('collapse',)
                            }),


        (_('Date accessi sistema'), {'fields': (('date_joined',
                                                 'last_login', ))
                                     }),
    )
    add_fieldsets = (
        (None, {
            'classes': ('wide',),
            'fields': ('username', 'password1', 'password2'),
        }),
    )


admin.site.unregister(User)
admin.site.register(User, CustomUserAdmin)


@admin.register(PersistentId)
class PersistentIdAdmin(admin.ModelAdmin):
    list_display = ('user', 'persistent_id', 'recipient_id', 'created')
    list_filter = ('created',)



from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.contrib.auth.models import AbstractUser
from django.contrib.auth import get_user_model


class User(AbstractUser):
    GENDER = (
        ('male', _('Maschio')),
        ('female', _('Femmina')),
        ('other', _('Altro')),
    )

    first_name = models.CharField(_('Name'), max_length=30,
                                  blank=True, null=True)
    last_name = models.CharField(_('Surname'), max_length=30,
                                 blank=True, null=True)
    is_active = models.BooleanField(_('active'), default=True)
    email = models.EmailField('email address', blank=True, null=True)
    taxpayer_id = models.CharField(_("Taxpayer's identification number"),
                                   max_length=32,
                                   blank=True, null=True)
    gender = models.CharField(_('Genere'), choices=GENDER,
                              max_length=12, blank=True, null=True)
    place_of_birth = models.CharField('Luogo di nascita', max_length=56,
                                      blank=True, null=True)
    birth_date = models.DateField('Data di nascita',
                                  null=True, blank=True)
    origin = models.CharField(_('from which conenctor this user come from'),
                              max_length=254,
                              blank=True, null=True)

    class Meta:
        ordering = ['username']
        verbose_name_plural = _("Users")

    @property
    def uid(self):
        return self.username.split('@')[0]

    def persistent_id(self, entityid):
        """ returns persistent id related to a recipient (sp) entity id
        """
        pid = PersistentId.objects.filter(user=self,
                                          recipient_id=entityid).last()
        if pid:
            return pid.persistent_id

    def get_oidc_birthdate(self):
        return self.birth_date.strftime('%Y-%m-%d') if self.birth_date else ''

    def get_oidc_lastlogin(self):
        return self.last_login.timestamp() if self.last_login else ''

    def __str__(self):
        return '{} {}'.format(self.first_name, self.last_name)


class PersistentId(models.Model):
    user = models.ForeignKey(get_user_model(), on_delete=models.CASCADE)
    persistent_id = models.CharField(_('Persistent Stored ID'),
                                     max_length=254,
                                     blank=True, null=True)
    recipient_id = models.CharField(_('Relying-Party entityID'),
                                    max_length=254,
                                    blank=True, null=True)
    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        verbose_name = _('Persistent Id')
        verbose_name_plural = _('Persistent Id')

    def __str__(self):
        return '{}: {} to {} [{}]'.format(self.user,
                                          self.persistent_id,
                                          self.recipient_id,
                                          self.created)



import datetime
import json
import pytz

from oidcmsg.message import Message
from cryptojwt.key_jar import KeyJar

from . application import oidcop_app


def dt2timestamp(value):
    return int(datetime.datetime.timestamp(value))


def timestamp2dt(value):
    pytz.utc.localize(datetime.datetime.fromtimestamp(value))


def aware_dt_from_timestamp(timestamp):
    dt = datetime.datetime.fromtimestamp(timestamp)
    return pytz.timezone("UTC").localize(dt, is_dst=None)


def decode_token(txt, attr_name='access_token', verify_sign=True):
    issuer = oidcop_app.srv_config['issuer']
    jwks_path = oidcop_app.srv_config['keys']['private_path']
    jwks = json.loads(open(jwks_path).read())

    key_jar = KeyJar()
    key_jar.import_jwks(jwks, issuer=issuer)

    msg = Message().from_jwt(txt,
                             keyjar=key_jar,
                             verify=verify_sign)
    return msg



import logging
import os
import json

from django.conf import settings
from oidcop.server import Server

from urllib.parse import urlparse
from oidcop.configure import OPConfiguration

folder = os.path.dirname(os.path.realpath(__file__))
logger = logging.getLogger(__name__)


def init_oidc_op_endpoints(app):
    op_config = app.srv_config

    op_config['issuer']
    server = Server(op_config, cwd=folder)

    for endp in server.endpoint.values():
        p = urlparse(endp.endpoint_path)
        _vpath = p.path.split('/')
        endp.vpath = _vpath

    return server


def oidc_provider_init_app(config, name='oidc_op', **kwargs):
    name = name or __name__
    app = type('OIDCAppEndpoint', (object,), {"srv_config": config})
    # Initialize the oidc_provider after views to be able to set correct urls
    app.endpoint_context = init_oidc_op_endpoints(app)
    return app


def oidcop_application(conf=None):
    conf = conf or settings.OIDCOP_CONFIG
    domain = getattr(settings, 'DOMAIN', None)
    port = getattr(settings, 'PORT', None)
    config = OPConfiguration(conf=conf['op']['server_info'],
                             domain=domain,
                             port=port)
    app = oidc_provider_init_app(config)
    os.environ['OIDCOP_CONFIG'] = json.dumps(conf)
    return app


oidcop_app = oidcop_application(conf=settings.OIDCOP_CONFIG)



# Generated by Django 3.1.1 on 2021-06-08 17:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('oidc_provider', '0010_remove_oidcsession_session_info'),
    ]

    operations = [
        migrations.AlterField(
            model_name='oidcissuedtoken',
            name='type',
            field=models.CharField(choices=[('authorization_code', 'authorization_code'), (
                'access_token', 'access_token'), ('id_token', 'id_token'), ('refresh_token', 'refresh_token')], max_length=32),
        ),
    ]



import json
import logging

from django.http import JsonResponse
from oidcop.exception import InvalidClient

from . application import oidcop_app
from . models import get_client_by_id

logger = logging.getLogger(__name__)


def debug_request(endpoint_name, request):
    _post = (f'POST:{request.POST}' if request.POST
             else f'BODY:{request.body.decode()}' if request.body
             else '')

    bearer = request.META.get('HTTP_AUTHORIZATION')
    _authz_header = f'- Authorization Header: {bearer}' if bearer else ''

    logger.debug(
        f'{endpoint_name} request GET: {request.GET} - {_post}{_authz_header}'
    )


def _fill_cdb(request) -> None:
    client_id = request.GET.get('client_id') or request.POST.get('client_id')
    _msg = f'Client {client_id} not found!'
    if client_id:
        client = get_client_by_id(client_id)
        if client:
            ec = oidcop_app.endpoint_context
            ec.endpoint_context.cdb = {
                client_id: client.serialize()
            }
            return
    else:
        logger.warning(_msg)
        raise InvalidClient(_msg)


def prepare_oidc_endpoint(func_to_decorate):
    """ store_params_in_session as a funcion decorator
    """
    def new_func(*original_args, **original_kwargs):
        request = original_args[0]
        _name = func_to_decorate.__name__
        debug_request(f'{_name}', request)

        ec = oidcop_app.endpoint_context
        # yes, in flush we believe ...
        ec.endpoint_context.session_manager.flush()

        return func_to_decorate(*original_args, **original_kwargs)

    return new_func


def fill_cdb_by_request(func_to_decorate):
    """ store_params_in_session as a funcion decorator
    """
    def new_func(*original_args, **original_kwargs):
        request = original_args[0]

        try:
            _fill_cdb(request)
            return func_to_decorate(*original_args, **original_kwargs)
        except InvalidClient as e:
            return JsonResponse(json.dumps({
                'error': 'invalid_request',
                'error_description': str(e),
                'method': request.method
            }), safe=False, status=403)
    return new_func



import logging
import random
import string

from django.test import TestCase
from django.utils import timezone
from oidc_provider.models import OidcRelyingParty


logger = logging.getLogger('oidc_provider')


def randomString(stringLength=10):
    """Generate a random string of fixed length """
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(stringLength))


CLIENT_ID = randomString()
CLIENT_TEST = {
    'client_id': '{}'.format(CLIENT_ID),
    'client_salt': '6flfsj0Z',
    'registration_access_token': 'z3PCMmC1HZ1QmXeXGOQMJpWQNQynM4xY',
    'registration_client_uri': f'https://127.0.0.1:8000/registration_api?client_id={CLIENT_ID}',
    'client_id_issued_at': 1575460012,
    'client_secret': '19cc69b70d0108f630e52f72f7a3bd37ba4e11678ad1a7434e9818e1',
    'client_secret_expires_at': 1575892012,
    'application_type': 'web',
    'contacts': ['ops@example.com'],
    'token_endpoint_auth_method': 'client_secret_basic',
    'jwks_uri': 'https://127.0.0.1:8099/static/jwks.json',
    'redirect_uris': [('https://127.0.0.1:8099/authz_cb/django_oidc_op', {})],
    'post_logout_redirect_uris': [('https://127.0.0.1:8099', None)],
    'response_types': ['code'],
    'grant_types': ['authorization_code']
}

CDB = {CLIENT_ID: CLIENT_TEST}


class TestOidcRelyingParty(TestCase):
    rp = randomString().upper()

    def test_create_rp(self):
        now = timezone.localtime()
        self.client = OidcRelyingParty.objects.create(
            client_id=self.rp,
            client_secret_expires_at=now,
            client_id_issued_at=now,
            is_active=True
        )
        logger.info('Created and fetched RP: {}'.format(self.client))

    def test_import_from_cdb(self):
        OidcRelyingParty.import_from_cdb(CDB)





# from dal import autocomplete
#
# class UserAutocomplete(autocomplete.Select2QuerySetView):
# def get_queryset(self):
# if not self.request.user.is_authenticated():
# return User.objects.none()
# qs = User.objects.all()
# if self.q:
# qs = qs.filter(
# username__icontains=self.q
# )
# return qs



# Generated by Django 3.1.1 on 2021-05-16 16:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('oidc_provider', '0007_auto_20210516_1230'),
    ]

    operations = [
        migrations.AlterField(
            model_name='oidcsession',
            name='sid',
            field=models.CharField(
                blank=True, max_length=254, null=True, unique=True),
        ),
    ]



# Generated by Django 3.1.1 on 2021-05-17 18:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('oidc_provider', '0008_auto_20210516_1653'),
    ]

    operations = [
        migrations.AlterField(
            model_name='oidcrpredirecturi',
            name='type',
            field=models.CharField(choices=[('redirect_uris', 'redirect_uris'), ('post_logout_redirect_uris', 'post_logout_redirect_uris'), (
                'frontchannel_logout_uri', 'frontchannel_logout_uri'), ('backchannel_logout_uri', 'backchannel_logout_uri')], max_length=33),
        ),
    ]



import datetime
import json
import logging
import pytz

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models
from django.utils import timezone


from . exceptions import InconsinstentSessionDump
from . utils import aware_dt_from_timestamp


logger = logging.getLogger(__name__)


OIDC_RESPONSE_TYPES = settings.OIDCOP_CONFIG['op']['server_info'][
    'endpoint']['authorization']['kwargs']['response_types_supported']

OIDC_TOKEN_AUTHN_METHODS = settings.OIDCOP_CONFIG['op']['server_info'][
    'endpoint']['token']['kwargs']['client_authn_method']

OIDC_GRANT_TYPES = settings.OIDCOP_CONFIG[
    'op']['server_info']['capabilities']['grant_types_supported']

TIMESTAMP_FIELDS = ['client_id_issued_at', 'client_secret_expires_at']


def get_client_by_id(client_id):
    client = OidcRelyingParty.objects.filter(
        client_id=client_id,
        is_active=True,
        client_secret_expires_at__gte=timezone.localtime()
    )
    if client:
        return client.first()


class TimeStampedModel(models.Model):
    """
    An abstract base class model that provides self-updating
    ``created`` and ``modified`` fields.
    """
    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True, editable=False)

    class Meta:
        abstract = True


class OidcRelyingParty(TimeStampedModel):
    """
    See: https://openid.net/specs/openid-connect-registration-1_0.html#ClientMetadata

    {
        'client_id': '86M1io6O2Vdy',
        'client_salt': 'ehXmVjYE',
        'registration_access_token': 'lRail9TKK3Cj4kZdSt3KDorKVxyQvVGL',
        'registration_client_uri': 'https://127.0.0.1:5000/registration_api?client_id=86M1io6O2Vdy',
        'client_id_issued_at': 1619384394,
        'client_secret': '9f9a5b6dc23daca606c3766a1c6a0de29a2009b007be3d1da7ff8ca5',
        'client_secret_expires_at': 1621976394,
        'application_type': 'web',
        'response_types': ['code'],
        'contacts': ['ops@example.com'],
        'token_endpoint_auth_method': 'client_secret_basic',
        'post_logout_redirect_uris': [('https://127.0.0.1:8090/session_logout/local', '')],
        'jwks_uri': 'https://127.0.0.1:8090/static/jwks.json',
        'frontchannel_logout_uri': 'https://127.0.0.1:8090/fc_logout/local',
        'frontchannel_logout_session_required': True,
        'backchannel_logout_uri': 'https://127.0.0.1:8090/bc_logout/local',
        'grant_types': ['authorization_code'],
        'redirect_uris': [('https://127.0.0.1:8090/authz_cb/local', {})]
    }


    """
    _TOKEN_AUTH_CHOICES = ((i, i) for i in OIDC_TOKEN_AUTHN_METHODS)

    client_id = models.CharField(
        max_length=255, blank=False, null=False, unique=True
    )
    client_salt = models.CharField(
        max_length=255, blank=True, null=True
    )
    registration_access_token = models.CharField(
        max_length=255, blank=True, null=True
    )
    registration_client_uri = models.URLField(
        max_length=255, blank=True, null=True
    )
    client_id_issued_at = models.DateTimeField(blank=True, null=True)
    client_secret = models.CharField(
        max_length=255,
        blank=True, null=True,
        help_text=('It is not needed for Clients '
                   'selecting a token_endpoint_auth_method '
                   'of private_key_jwt')
    )
    client_secret_expires_at = models.DateTimeField(
        blank=True, null=True,
        help_text=('REQUIRED if client_secret is issued')
    )
    application_type = models.CharField(
        max_length=255, blank=True,
        null=True, default='web'
    )
    token_endpoint_auth_method = models.CharField(
        choices=_TOKEN_AUTH_CHOICES,
        max_length=33,
        blank=True, null=True,
        default="client_secret_basic"
    )
    jwks_uri = models.URLField(max_length=255, blank=True, null=True)
    is_active = models.BooleanField(('active'), default=True)
    last_seen = models.DateTimeField(blank=True, null=True)

    @property
    def allowed_scopes(self):
        scopes = self.oidcrpscope_set.filter(client=self)
        return [i.scope for i in scopes] if scopes else None

    @allowed_scopes.setter
    def allowed_scopes(self, values):
        if not values:
            return None
        for i in values:
            data = dict(client=self, scope=i)
            if not self.oidcrpscope_set.filter(**data):
                self.oidcrpscope_set.create(**data)

    @property
    def contacts(self):
        return [
            elem.contact
            for elem in self.oidcrpcontact_set.filter(client=self)
        ]

    @contacts.setter
    def contacts(self, values):
        self.oidcrpcontact_set.filter(client=self).delete()
        if isinstance(values, str):
            value = [values]
        for value in values:
            data = dict(client=self, contact=value)
            if not self.oidcrpcontact_set.filter(**data):
                self.oidcrpcontact_set.create(**data)

    @property
    def grant_types(self):
        return [
            elem.grant_type
            for elem in
            self.oidcrpgranttype_set.filter(client=self)
        ]

    @grant_types.setter
    def grant_types(self, values):
        self.oidcrpgranttype_set.filter(client=self).delete()
        if isinstance(values, str):
            value = [values]
        for value in values:
            data = dict(client=self, grant_type=value)
            if not self.oidcrpgranttype_set.filter(**data):
                self.oidcrpgranttype_set.create(**data)

    @property
    def response_types(self):
        return [
            elem.response_type
            for elem in
            self.oidcrpresponsetype_set.filter(client=self)
        ]

    @response_types.setter
    def response_types(self, values):
        self.oidcrpresponsetype_set.filter(client=self).delete()
        if isinstance(values, str):
            value = [values]
        for value in values:
            data = dict(client=self, response_type=value)
            if not self.oidcrpresponsetype_set.filter(**data):
                self.oidcrpresponsetype_set.create(**data)

    def get_redirect_uri(self, uri_type: str):
        res = []
        for elem in self.oidcrpredirecturi_set.\
                filter(client=self, type=uri_type):
            res.append((elem.uri, json.loads(elem.values)))
        return res

    def set_redirect_uri(self, uri_type: str, values):
        self.oidcrpredirecturi_set.filter(client=self, type=uri_type).delete()
        for value in values:
            args = json.dumps(value[1] if value[1] else [])
            data = dict(
                client=self,
                uri=value[0],
                values=args,
                type=uri_type
            )
            self.oidcrpredirecturi_set.create(**data)

    def save(self, *args, **kwargs):
        for d in TIMESTAMP_FIELDS:
            field = getattr(self, d)
            if d == 'client_id_issued_at' and not field:
                setattr(self, d, timezone.localtime())
            elif d == 'client_secret_expires_at' and not field:
                setattr(
                    self, d,
                    timezone.localtime() + timezone.timedelta(days=365)
                )

            field = getattr(self, d)
            if field.tzinfo is None or field.tzinfo.utcoffset(d) is None:
                timezone.activate(pytz.timezone("UTC"))
                setattr(self, d, timezone.make_aware(field))

        super().save(*args, **kwargs)

    @property
    def post_logout_redirect_uris(self):
        return self.get_redirect_uri('post_logout_redirect_uris')

    @post_logout_redirect_uris.setter
    def post_logout_redirect_uris(self, values):
        self.set_redirect_uri('post_logout_redirect_uris', values)

    @property
    def redirect_uris(self):
        return self.get_redirect_uri('redirect_uris')

    @redirect_uris.setter
    def redirect_uris(self, values):
        self.set_redirect_uri('redirect_uris', values)

    class Meta:
        verbose_name = ('Relying Party')
        verbose_name_plural = ('Relying Parties')

    @classmethod
    def import_from_cdb(cls, cdb):
        clients = []
        for client_id in cdb:
            if cls.objects.filter(client_id=client_id):  # pragma: no cover
                continue
            client = cls.objects.create(client_id=client_id)
            for k, v in cdb[client_id].items():
                if k in ('client_secret_expires_at', 'client_id_issued_at'):
                    if v:
                        v = datetime.datetime.fromtimestamp(v)
                    else:
                        v = timezone.localtime() + timezone.timedelta(days=1)
                setattr(client, k, v)
            client.save()
            clients.append(client)
        return clients

    def serialize(self):
        """
            Compability with rohe approach based on dictionaries
        """
        data = {k: v for k, v in self.__dict__.items() if k[0] != '_'}
        disabled = ('created', 'modified', 'is_active', 'last_seen', 'id')
        for dis in disabled:
            data.pop(dis)
        for key in TIMESTAMP_FIELDS:
            if data.get(key):
                data[key] = int(datetime.datetime.timestamp(data[key]))

        for i in ('contacts', 'grant_types', 'response_types',
                  'post_logout_redirect_uris', 'redirect_uris'):
            data[i] = getattr(self, i)

        if self.allowed_scopes:
            # without allowed scopes set it will return all availables
            # configure allowed_scopes to filter only which one MUST allowed
            data['allowed_scopes'] = self.allowed_scopes

        return data

    def __str__(self):
        return '{}'.format(self.client_id)


class OidcRPResponseType(TimeStampedModel):
    client = models.ForeignKey(OidcRelyingParty, on_delete=models.CASCADE)
    response_type = models.CharField(choices=[
        (i, i) for i in OIDC_RESPONSE_TYPES
    ],
        max_length=60
    )

    class Meta:
        verbose_name = ('Relying Party Response Type')
        verbose_name_plural = ('Relying Parties Response Types')
        unique_together = ('client', 'response_type')

    def __str__(self):
        return '{}, [{}]'.format(self.client, self.response_type)


class OidcRPGrantType(TimeStampedModel):
    client = models.ForeignKey(OidcRelyingParty,
                               on_delete=models.CASCADE)
    grant_type = models.CharField(choices=[(i, i)
                                           for i in OIDC_GRANT_TYPES],
                                  max_length=60)

    class Meta:
        verbose_name = ('Relying Party GrantType')
        verbose_name_plural = ('Relying Parties GrantTypes')
        unique_together = ('client', 'grant_type')

    def __str__(self):
        return '{}, [{}]'.format(self.client, self.grant_type)


class OidcRPContact(TimeStampedModel):
    client = models.ForeignKey(OidcRelyingParty,
                               on_delete=models.CASCADE)
    contact = models.CharField(max_length=254,
                               blank=True, null=True,)

    class Meta:
        verbose_name = ('Relying Party Contact')
        verbose_name_plural = ('Relying Parties Contacts')
        unique_together = ('client', 'contact')

    def __str__(self):
        return '{}, [{}]'.format(self.client, self.contact)


class OidcRPRedirectUri(TimeStampedModel):
    URIS = (
        ('redirect_uris', 'redirect_uris'),
        ('post_logout_redirect_uris', 'post_logout_redirect_uris'),
        ('frontchannel_logout_uri', 'frontchannel_logout_uri'),
        ('backchannel_logout_uri', 'backchannel_logout_uri'),
    )

    client = models.ForeignKey(OidcRelyingParty,
                               on_delete=models.CASCADE)
    uri = models.CharField(max_length=254,
                           blank=True, null=True)
    values = models.CharField(max_length=254,
                              blank=True, null=True)
    type = models.CharField(choices=URIS,
                            max_length=33)

    class Meta:
        verbose_name = ('Relying Party URI')
        verbose_name_plural = ('Relying Parties URIs')

    def __str__(self):
        return '{} [{}] {}'.format(self.client, self.uri, self.type)


class OidcRPScope(TimeStampedModel):
    client = models.ForeignKey(OidcRelyingParty,
                               on_delete=models.CASCADE)
    scope = models.CharField(max_length=254,
                             blank=True, null=True,)

    class Meta:
        verbose_name = ('Relying Party Scope')
        verbose_name_plural = ('Relying Parties Scopes')
        unique_together = ('client', 'scope')

    def __str__(self):
        return '{}, [{}]'.format(self.client, self.scope)


class OidcSession(TimeStampedModel):
    """
    Store UserSessionInfo, ClientSessionInfo and Grant
    """

    user_uid = models.CharField(max_length=120, blank=True, null=True)
    user = models.ForeignKey(get_user_model(), on_delete=models.CASCADE,
                             blank=True, null=True)
    client = models.ForeignKey(OidcRelyingParty, on_delete=models.CASCADE,
                               blank=True, null=True)

    user_sessioninfo = models.TextField(default='{}')
    client_sessioninfo = models.TextField(default='{}')

    grant_sessioninfo = models.TextField()
    grant_uid = models.CharField(max_length=254, blank=True, null=True)
    expires_at = models.DateTimeField(blank=True, null=True)
    revoked = models.BooleanField(default=False)

    sub = models.CharField(max_length=254, blank=True, null=True)
    sid = models.CharField(max_length=254, unique=True, blank=True, null=True)
    sid_encrypted = models.CharField(max_length=254, blank=True, null=True)

    key = models.CharField(max_length=254, blank=True, null=True)
    salt = models.CharField(max_length=254, blank=True, null=True)

    class Meta:
        verbose_name = ('SSO Session')
        verbose_name_plural = ('SSO Sessions')

    @property
    def user_session_info(self):
        return json.loads(self.user_sessioninfo)

    @property
    def client_session_info(self):
        return json.loads(self.client_sessioninfo)

    @property
    def grant(self):
        return json.loads(self.grant_sessioninfo)

    @classmethod
    def load(cls, ses_man_dump: dict) -> dict:
        if 'db' not in ses_man_dump:
            ses_man_dump

        attr_map = {
            'oidcop.session.info.UserSessionInfo': 'user_sessioninfo',
            'oidcop.session.info.ClientSessionInfo': 'client_sessioninfo',
            'oidcop.session.grant.Grant': 'grant_sessioninfo'
        }
        data = dict()
        for k, v in ses_man_dump['db'].items():

            # TODO: ask roland to have something more precise
            if len(k) > 128 and k not in attr_map.values():
                data['sid_encrypted'] = k
                continue

            classname = v[0]
            attr = getattr(cls, attr_map[classname])
            field_name = attr.field.name
            data[field_name] = json.dumps(v[1])
            if field_name == 'user_sessioninfo':
                user_id = v[1]['user_id']
                client_id = v[1]['subordinate'][0]
                data['user_uid'] = user_id
                data['user'] = get_user_model().objects.get(username=user_id)
                data['client'] = OidcRelyingParty.objects.get(
                    client_id=client_id)
            elif field_name == 'client_sessioninfo':
                data['grant_uid'] = v[1]['subordinate'][0]
            elif field_name == 'grant_sessioninfo':
                data['expires_at'] = aware_dt_from_timestamp(
                    v[1]['expires_at'])
                data['revoked'] = v[1]['revoked']
                data['sub'] = v[1]['sub']
                data['sid'] = f"{user_id};;{client_id};;{data['grant_uid']}"

        data['key'] = ses_man_dump['key']
        data['salt'] = ses_man_dump['salt']

        if not ses_man_dump['key']:
            logger.critical(
                "Missing key in session dump"
            )

        if not ses_man_dump['salt']:
            logger.critical(
                "Missing salt in session dump"
            )

        session = cls.objects.filter(sid=data['sid'])
        if not session:
            session = cls.objects.create(**data)
        else:
            session.update(**data)
            session = session.first()
        OidcIssuedToken.load(session)

        if session.serialize() != ses_man_dump:
            logger.critical(ses_man_dump, session)
            raise InconsinstentSessionDump(
                'Serialized session differs from the dumped one'
            )
        return session

    def serialize(self):
        user_label = self.user_uid
        ses_label = f"{user_label};;{self.client.client_id}"
        grant_label = f"{ses_label};;{self.grant_uid}"

        return dict(
            db={
                user_label: [
                    'oidcop.session.info.UserSessionInfo',
                    self.user_session_info
                ],
                ses_label: [
                    'oidcop.session.info.ClientSessionInfo',
                    self.client_session_info
                ],
                grant_label: [
                    'oidcop.session.grant.Grant',
                    self.grant
                ],
                self.sid_encrypted: [
                    'oidcop.session.grant.Grant', self.grant]
            },
            salt=self.salt,
            key=self.key
        )

    def __str__(self):
        return f'{self.user.username};;{self.client.client_id};;{self.grant_uid}'


class OidcIssuedToken(TimeStampedModel):
    """
    Stores issued token
    """
    TT_CHOICES = (
        ('authorization_code', 'authorization_code'),
        ('access_token', 'access_token'),
        ('id_token', 'id_token'),
        ('refresh_token', 'refresh_token')
    )

    uid = models.CharField(max_length=128, blank=True, null=True)
    type = models.CharField(choices=TT_CHOICES,
                            max_length=32,
                            blank=False, null=False)

    issued_at = models.DateTimeField()
    expires_at = models.DateTimeField()
    not_before = models.DateTimeField(null=True, blank=True)

    revoked = models.BooleanField(default=False)
    value = models.TextField(unique=True)

    usage_rules = models.TextField()
    used = models.IntegerField(default=0)
    based_on = models.TextField(blank=True, null=True)

    session = models.ForeignKey(OidcSession, on_delete=models.CASCADE)

    class Meta:
        verbose_name = ('Issued Token')
        verbose_name_plural = ('Issued Tokens')
        indexes = [
            models.Index(fields=['value', ]),
        ]

    def serialize(self):
        return {
            "token_class": self.type,
            "issued_at": self.issued_at.timestamp(),
            "expires_at": self.expires_at.timestamp(),
            "not_before": self.not_before.timestamp() if self.not_before else 0,
            "revoked": self.revoked,
            "value": self.value,
            "usage_rules": self.usage_rules,
            "used": self.used,
            "based_on": self.based_on,
            "id": self.uid
        }

    @classmethod
    def load(cls, session: OidcSession) -> None:
        for token in session.grant['issued_token']:
            token = token[list(token.keys())[0]]

            if token.get('not_before'):
                nbt = datetime.datetime.fromtimestamp(token['not_before'])
            else:
                nbt = None

            data = dict(
                session=session,
                type=token['token_class'],
                issued_at=aware_dt_from_timestamp(token['issued_at']),
                expires_at=aware_dt_from_timestamp(token['expires_at']),
                not_before=nbt,
                revoked=token['revoked'],
                value=token['value'],
                usage_rules=json.dumps(token['usage_rules'],),
                used=token['used'],
                based_on=token.get('based_on'),
                uid=token['id'],
            )

            obj = cls.objects.filter(
                session=session, value=data['value']
            )
            if not obj:
                obj = cls.objects.create(**data)
            else:
                obj.update(**data)

    def __str__(self):
        return f'{self.type} {self.session}'



# Generated by Django 3.1.1 on 2021-05-10 12:24

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='OidcRelyingParty',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('client_id', models.CharField(max_length=255, unique=True)),
                ('client_salt', models.CharField(
                    blank=True, max_length=255, null=True)),
                ('registration_access_token', models.CharField(
                    blank=True, max_length=255, null=True)),
                ('registration_client_uri', models.URLField(
                    blank=True, max_length=255, null=True)),
                ('client_id_issued_at', models.DateTimeField(blank=True, null=True)),
                ('client_secret', models.CharField(
                    blank=True, help_text='It is not needed for Clients selecting a token_endpoint_auth_method of private_key_jwt', max_length=255, null=True)),
                ('client_secret_expires_at', models.DateTimeField(blank=True,
                                                                  help_text='REQUIRED if client_secret is issued', null=True)),
                ('application_type', models.CharField(
                    blank=True, default='web', max_length=255, null=True)),
                ('token_endpoint_auth_method', models.CharField(blank=True, choices=[('client_secret_post', 'client_secret_post'), ('client_secret_basic', 'client_secret_basic'), (
                    'client_secret_jwt', 'client_secret_jwt'), ('private_key_jwt', 'private_key_jwt')], default='client_secret_basic', max_length=33, null=True)),
                ('jwks_uri', models.URLField(blank=True, max_length=255, null=True)),
                ('is_active', models.BooleanField(
                    default=True, verbose_name='active')),
                ('last_seen', models.DateTimeField(blank=True, null=True)),
            ],
            options={
                'verbose_name': 'Relying Party',
                'verbose_name_plural': 'Relying Parties',
            },
        ),
        migrations.CreateModel(
            name='OidcSession',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('user_uid', models.CharField(blank=True, max_length=120, null=True)),
                ('user_sessioninfo', models.TextField(default='{}')),
                ('client_sessioninfo', models.TextField(default='{}')),
                ('grant_sessioninfo', models.TextField()),
                ('grant_uid', models.CharField(
                    blank=True, max_length=254, null=True)),
                ('sub', models.CharField(blank=True, max_length=255, null=True)),
                ('expires_at', models.DateTimeField(blank=True, null=True)),
                ('revoked', models.BooleanField(default=False)),
                ('session_info', models.TextField(
                    blank=True, default='{}', null=True)),
                ('sid', models.CharField(blank=True, max_length=255, null=True)),
                ('client', models.ForeignKey(blank=True, null=True,
                                             on_delete=django.db.models.deletion.CASCADE, to='oidc_provider.oidcrelyingparty')),
                ('user', models.ForeignKey(blank=True, null=True,
                                           on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'SSO Session',
                'verbose_name_plural': 'SSO Sessions',
            },
        ),
        migrations.CreateModel(
            name='OidcRPRedirectUri',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('uri', models.CharField(blank=True, max_length=254, null=True)),
                ('values', models.CharField(blank=True, max_length=254, null=True)),
                ('type', models.CharField(choices=[('redirect_uris', 'redirect_uris'), (
                    'post_logout_redirect_uris', 'post_logout_redirect_uris')], max_length=33)),
                ('client', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='oidc_provider.oidcrelyingparty')),
            ],
            options={
                'verbose_name': 'Relying Party URI',
                'verbose_name_plural': 'Relying Parties URIs',
            },
        ),
        migrations.CreateModel(
            name='OidcIssuedToken',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('uid', models.CharField(blank=True, max_length=128, null=True)),
                ('type', models.CharField(choices=[('authorization_code', 'authorization_code'), (
                    'access_token', 'access_token'), ('id_token', 'id_token')], max_length=32)),
                ('issued_at', models.DateTimeField()),
                ('expires_at', models.DateTimeField()),
                ('not_before', models.DateTimeField(blank=True, null=True)),
                ('revoked', models.BooleanField(default=False)),
                ('value', models.TextField()),
                ('usage_rules', models.TextField()),
                ('used', models.IntegerField(default=0)),
                ('based_on', models.TextField(blank=True, null=True)),
                ('session', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='oidc_provider.oidcsession')),
            ],
            options={
                'verbose_name': 'OIDC Issued Token',
                'verbose_name_plural': 'OIDC Issued Tokens',
            },
        ),
        migrations.CreateModel(
            name='OidcRPScope',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('scope', models.CharField(blank=True, max_length=254, null=True)),
                ('client', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='oidc_provider.oidcrelyingparty')),
            ],
            options={
                'verbose_name': 'Relying Party Scope',
                'verbose_name_plural': 'Relying Parties Scopes',
                'unique_together': {('client', 'scope')},
            },
        ),
        migrations.CreateModel(
            name='OidcRPResponseType',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('response_type', models.CharField(
                    choices=[('code', 'code')], max_length=60)),
                ('client', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='oidc_provider.oidcrelyingparty')),
            ],
            options={
                'verbose_name': 'Relying Party Response Type',
                'verbose_name_plural': 'Relying Parties Response Types',
                'unique_together': {('client', 'response_type')},
            },
        ),
        migrations.CreateModel(
            name='OidcRPGrantType',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('grant_type', models.CharField(choices=[('authorization_code', 'authorization_code'), ('implicit', 'implicit'), (
                    'urn:ietf:params:oauth:grant-type:jwt-bearer', 'urn:ietf:params:oauth:grant-type:jwt-bearer'), ('refresh_token', 'refresh_token')], max_length=60)),
                ('client', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='oidc_provider.oidcrelyingparty')),
            ],
            options={
                'verbose_name': 'Relying Party GrantType',
                'verbose_name_plural': 'Relying Parties GrantTypes',
                'unique_together': {('client', 'grant_type')},
            },
        ),
        migrations.CreateModel(
            name='OidcRPContact',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('contact', models.CharField(blank=True, max_length=254, null=True)),
                ('client', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='oidc_provider.oidcrelyingparty')),
            ],
            options={
                'verbose_name': 'Relying Party Contact',
                'verbose_name_plural': 'Relying Parties Contacts',
                'unique_together': {('client', 'contact')},
            },
        ),
        migrations.AddIndex(
            model_name='oidcissuedtoken',
            index=models.Index(
                fields=['value'], name='oidc_provid_value_6594c4_idx'),
        ),
    ]



# Generated by Django 2.2.5 on 2019-09-21 16:47

from django.conf import settings
import django.contrib.auth.models
import django.contrib.auth.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0011_update_proxy_permissions'),
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('password', models.CharField(
                    max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(
                    blank=True, null=True, verbose_name='last login')),
                ('is_superuser', models.BooleanField(default=False,
                                                     help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='superuser status')),
                ('username', models.CharField(error_messages={'unique': 'A user with that username already exists.'}, help_text='Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.',
                                              max_length=150, unique=True, validators=[django.contrib.auth.validators.UnicodeUsernameValidator()], verbose_name='username')),
                ('is_staff', models.BooleanField(default=False,
                                                 help_text='Designates whether the user can log into this admin site.', verbose_name='staff status')),
                ('date_joined', models.DateTimeField(
                    default=django.utils.timezone.now, verbose_name='date joined')),
                ('first_name', models.CharField(blank=True,
                                                max_length=30, null=True, verbose_name='Name')),
                ('last_name', models.CharField(blank=True,
                                               max_length=30, null=True, verbose_name='Surname')),
                ('is_active', models.BooleanField(
                    default=True, verbose_name='active')),
                ('email', models.EmailField(blank=True, max_length=254,
                                            null=True, verbose_name='email address')),
                ('taxpayer_id', models.CharField(blank=True, max_length=32,
                                                 null=True, verbose_name="Taxpayer's identification number")),
                ('gender', models.CharField(blank=True, choices=[('male', 'Maschio'), ('female', 'Femmina'), (
                    'other', 'Altro')], max_length=12, null=True, verbose_name='Genere')),
                ('place_of_birth', models.CharField(blank=True, choices=[('Aruba', 'Aruba'), ('Afghanistan', 'Afghanistan'), ('Angola', 'Angola'), ('Anguilla', 'Anguilla'), ('Åland Islands', 'Åland Islands'), ('Albania', 'Albania'), ('Andorra', 'Andorra'), ('United Arab Emirates', 'United Arab Emirates'), ('Argentina', 'Argentina'), ('Armenia', 'Armenia'), ('American Samoa', 'American Samoa'), ('Antarctica', 'Antarctica'), ('French Southern Territories', 'French Southern Territories'), ('Antigua and Barbuda', 'Antigua and Barbuda'), ('Australia', 'Australia'), ('Austria', 'Austria'), ('Azerbaijan', 'Azerbaijan'), ('Burundi', 'Burundi'), ('Belgium', 'Belgium'), ('Benin', 'Benin'), ('Bonaire, Sint Eustatius and Saba', 'Bonaire, Sint Eustatius and Saba'), ('Burkina Faso', 'Burkina Faso'), ('Bangladesh', 'Bangladesh'), ('Bulgaria', 'Bulgaria'), ('Bahrain', 'Bahrain'), ('Bahamas', 'Bahamas'), ('Bosnia and Herzegovina', 'Bosnia and Herzegovina'), ('Saint Barthélemy', 'Saint Barthélemy'), ('Belarus', 'Belarus'), ('Belize', 'Belize'), ('Bermuda', 'Bermuda'), ('Bolivia, Plurinational State of', 'Bolivia, Plurinational State of'), ('Brazil', 'Brazil'), ('Barbados', 'Barbados'), ('Brunei Darussalam', 'Brunei Darussalam'), ('Bhutan', 'Bhutan'), ('Bouvet Island', 'Bouvet Island'), ('Botswana', 'Botswana'), ('Central African Republic', 'Central African Republic'), ('Canada', 'Canada'), ('Cocos (Keeling) Islands', 'Cocos (Keeling) Islands'), ('Switzerland', 'Switzerland'), ('Chile', 'Chile'), ('China', 'China'), ("Côte d'Ivoire", "Côte d'Ivoire"), ('Cameroon', 'Cameroon'), ('Congo, The Democratic Republic of the', 'Congo, The Democratic Republic of the'), ('Congo', 'Congo'), ('Cook Islands', 'Cook Islands'), ('Colombia', 'Colombia'), ('Comoros', 'Comoros'), ('Cabo Verde', 'Cabo Verde'), ('Costa Rica', 'Costa Rica'), ('Cuba', 'Cuba'), ('Curaçao', 'Curaçao'), ('Christmas Island', 'Christmas Island'), ('Cayman Islands', 'Cayman Islands'), ('Cyprus', 'Cyprus'), ('Czechia', 'Czechia'), ('Germany', 'Germany'), ('Djibouti', 'Djibouti'), ('Dominica', 'Dominica'), ('Denmark', 'Denmark'), ('Dominican Republic', 'Dominican Republic'), ('Algeria', 'Algeria'), ('Ecuador', 'Ecuador'), ('Egypt', 'Egypt'), ('Eritrea', 'Eritrea'), ('Western Sahara', 'Western Sahara'), ('Spain', 'Spain'), ('Estonia', 'Estonia'), ('Ethiopia', 'Ethiopia'), ('Finland', 'Finland'), ('Fiji', 'Fiji'), ('Falkland Islands (Malvinas)', 'Falkland Islands (Malvinas)'), ('France', 'France'), ('Faroe Islands', 'Faroe Islands'), ('Micronesia, Federated States of', 'Micronesia, Federated States of'), ('Gabon', 'Gabon'), ('United Kingdom', 'United Kingdom'), ('Georgia', 'Georgia'), ('Guernsey', 'Guernsey'), ('Ghana', 'Ghana'), ('Gibraltar', 'Gibraltar'), ('Guinea', 'Guinea'), ('Guadeloupe', 'Guadeloupe'), ('Gambia', 'Gambia'), ('Guinea-Bissau', 'Guinea-Bissau'), ('Equatorial Guinea', 'Equatorial Guinea'), ('Greece', 'Greece'), ('Grenada', 'Grenada'), ('Greenland', 'Greenland'), ('Guatemala', 'Guatemala'), ('French Guiana', 'French Guiana'), ('Guam', 'Guam'), ('Guyana', 'Guyana'), ('Hong Kong', 'Hong Kong'), ('Heard Island and McDonald Islands', 'Heard Island and McDonald Islands'), ('Honduras', 'Honduras'), ('Croatia', 'Croatia'), ('Haiti', 'Haiti'), ('Hungary', 'Hungary'), ('Indonesia', 'Indonesia'), ('Isle of Man', 'Isle of Man'), ('India', 'India'), ('British Indian Ocean Territory', 'British Indian Ocean Territory'), ('Ireland', 'Ireland'), ('Iran, Islamic Republic of', 'Iran, Islamic Republic of'), ('Iraq', 'Iraq'), ('Iceland', 'Iceland'), ('Israel', 'Israel'), ('Italy', 'Italy'), ('Jamaica', 'Jamaica'), ('Jersey', 'Jersey'), ('Jordan', 'Jordan'), ('Japan', 'Japan'), ('Kazakhstan', 'Kazakhstan'), ('Kenya', 'Kenya'), ('Kyrgyzstan', 'Kyrgyzstan'), ('Cambodia', 'Cambodia'), ('Kiribati', 'Kiribati'), ('Saint Kitts and Nevis', 'Saint Kitts and Nevis'), ('Korea, Republic of', 'Korea, Republic of'), ('Kuwait', 'Kuwait'), ("Lao People's Democratic Republic", "Lao People's Democratic Republic"), ('Lebanon', 'Lebanon'), ('Liberia', 'Liberia'), ('Libya', 'Libya'), ('Saint Lucia', 'Saint Lucia'), (
                    'Liechtenstein', 'Liechtenstein'), ('Sri Lanka', 'Sri Lanka'), ('Lesotho', 'Lesotho'), ('Lithuania', 'Lithuania'), ('Luxembourg', 'Luxembourg'), ('Latvia', 'Latvia'), ('Macao', 'Macao'), ('Saint Martin (French part)', 'Saint Martin (French part)'), ('Morocco', 'Morocco'), ('Monaco', 'Monaco'), ('Moldova, Republic of', 'Moldova, Republic of'), ('Madagascar', 'Madagascar'), ('Maldives', 'Maldives'), ('Mexico', 'Mexico'), ('Marshall Islands', 'Marshall Islands'), ('North Macedonia', 'North Macedonia'), ('Mali', 'Mali'), ('Malta', 'Malta'), ('Myanmar', 'Myanmar'), ('Montenegro', 'Montenegro'), ('Mongolia', 'Mongolia'), ('Northern Mariana Islands', 'Northern Mariana Islands'), ('Mozambique', 'Mozambique'), ('Mauritania', 'Mauritania'), ('Montserrat', 'Montserrat'), ('Martinique', 'Martinique'), ('Mauritius', 'Mauritius'), ('Malawi', 'Malawi'), ('Malaysia', 'Malaysia'), ('Mayotte', 'Mayotte'), ('Namibia', 'Namibia'), ('New Caledonia', 'New Caledonia'), ('Niger', 'Niger'), ('Norfolk Island', 'Norfolk Island'), ('Nigeria', 'Nigeria'), ('Nicaragua', 'Nicaragua'), ('Niue', 'Niue'), ('Netherlands', 'Netherlands'), ('Norway', 'Norway'), ('Nepal', 'Nepal'), ('Nauru', 'Nauru'), ('New Zealand', 'New Zealand'), ('Oman', 'Oman'), ('Pakistan', 'Pakistan'), ('Panama', 'Panama'), ('Pitcairn', 'Pitcairn'), ('Peru', 'Peru'), ('Philippines', 'Philippines'), ('Palau', 'Palau'), ('Papua New Guinea', 'Papua New Guinea'), ('Poland', 'Poland'), ('Puerto Rico', 'Puerto Rico'), ("Korea, Democratic People's Republic of", "Korea, Democratic People's Republic of"), ('Portugal', 'Portugal'), ('Paraguay', 'Paraguay'), ('Palestine, State of', 'Palestine, State of'), ('French Polynesia', 'French Polynesia'), ('Qatar', 'Qatar'), ('Réunion', 'Réunion'), ('Romania', 'Romania'), ('Russian Federation', 'Russian Federation'), ('Rwanda', 'Rwanda'), ('Saudi Arabia', 'Saudi Arabia'), ('Sudan', 'Sudan'), ('Senegal', 'Senegal'), ('Singapore', 'Singapore'), ('South Georgia and the South Sandwich Islands', 'South Georgia and the South Sandwich Islands'), ('Saint Helena, Ascension and Tristan da Cunha', 'Saint Helena, Ascension and Tristan da Cunha'), ('Svalbard and Jan Mayen', 'Svalbard and Jan Mayen'), ('Solomon Islands', 'Solomon Islands'), ('Sierra Leone', 'Sierra Leone'), ('El Salvador', 'El Salvador'), ('San Marino', 'San Marino'), ('Somalia', 'Somalia'), ('Saint Pierre and Miquelon', 'Saint Pierre and Miquelon'), ('Serbia', 'Serbia'), ('South Sudan', 'South Sudan'), ('Sao Tome and Principe', 'Sao Tome and Principe'), ('Suriname', 'Suriname'), ('Slovakia', 'Slovakia'), ('Slovenia', 'Slovenia'), ('Sweden', 'Sweden'), ('Eswatini', 'Eswatini'), ('Sint Maarten (Dutch part)', 'Sint Maarten (Dutch part)'), ('Seychelles', 'Seychelles'), ('Syrian Arab Republic', 'Syrian Arab Republic'), ('Turks and Caicos Islands', 'Turks and Caicos Islands'), ('Chad', 'Chad'), ('Togo', 'Togo'), ('Thailand', 'Thailand'), ('Tajikistan', 'Tajikistan'), ('Tokelau', 'Tokelau'), ('Turkmenistan', 'Turkmenistan'), ('Timor-Leste', 'Timor-Leste'), ('Tonga', 'Tonga'), ('Trinidad and Tobago', 'Trinidad and Tobago'), ('Tunisia', 'Tunisia'), ('Turkey', 'Turkey'), ('Tuvalu', 'Tuvalu'), ('Taiwan, Province of China', 'Taiwan, Province of China'), ('Tanzania, United Republic of', 'Tanzania, United Republic of'), ('Uganda', 'Uganda'), ('Ukraine', 'Ukraine'), ('United States Minor Outlying Islands', 'United States Minor Outlying Islands'), ('Uruguay', 'Uruguay'), ('United States', 'United States'), ('Uzbekistan', 'Uzbekistan'), ('Holy See (Vatican City State)', 'Holy See (Vatican City State)'), ('Saint Vincent and the Grenadines', 'Saint Vincent and the Grenadines'), ('Venezuela, Bolivarian Republic of', 'Venezuela, Bolivarian Republic of'), ('Virgin Islands, British', 'Virgin Islands, British'), ('Virgin Islands, U.S.', 'Virgin Islands, U.S.'), ('Viet Nam', 'Viet Nam'), ('Vanuatu', 'Vanuatu'), ('Wallis and Futuna', 'Wallis and Futuna'), ('Samoa', 'Samoa'), ('Yemen', 'Yemen'), ('South Africa', 'South Africa'), ('Zambia', 'Zambia'), ('Zimbabwe', 'Zimbabwe')], max_length=30, null=True, verbose_name='Luogo di nascita')),
                ('birth_date', models.DateField(blank=True,
                                                null=True, verbose_name='Data di nascita')),
                ('origin', models.CharField(blank=True, max_length=254, null=True,
                                            verbose_name='from which conenctor this user come from')),
                ('groups', models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.',
                                                  related_name='user_set', related_query_name='user', to='auth.Group', verbose_name='groups')),
                ('user_permissions', models.ManyToManyField(blank=True, help_text='Specific permissions for this user.',
                                                            related_name='user_set', related_query_name='user', to='auth.Permission', verbose_name='user permissions')),
            ],
            options={
                'verbose_name_plural': 'Users',
                'ordering': ['username'],
            },
            managers=[
                ('objects', django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name='PersistentId',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('persistent_id', models.CharField(blank=True, max_length=254,
                                                   null=True, verbose_name='SAML Persistent Stored ID')),
                ('recipient_id', models.CharField(blank=True, max_length=254,
                                                  null=True, verbose_name='SAML ServiceProvider entityID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('user', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Persistent Id',
                'verbose_name_plural': 'Persistent Id',
            },
        ),
    ]



"""
Django settings for example project.

Generated by 'django-admin startproject' using Django 2.2.5.

For more information on this file, see
https://docs.djangoproject.com/en/2.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.2/ref/settings/
"""

import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'hm8_+266i)9@*9#ncep59m+mp$_+8d_#19rd5x0ujtwlp140s@'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ['*']

# COOKIES
SESSION_COOKIE_SECURE = True
SESSION_COOKIE_SAMESITE = "None"
SESSION_EXPIRE_AT_BROWSER_CLOSE = True
CSRF_COOKIE_HTTPONLY = True
CSRF_COOKIE_SAMESITE = "None"
CSRF_COOKIE_SECURE = True

# Application definition

INSTALLED_APPS = [
    'accounts',

    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    'oidc_provider',

]

if 'oidc_provider' in INSTALLED_APPS:
    from . oidc_provider_settings import DOMAIN, OIDCOP_CONFIG, OIDC_URL_PREFIX, PORT
    OIDCOP_CONF_FILE = f"{BASE_DIR}/oidc_provider_settings.py"


MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'example.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'example.wsgi.application'


# Database
# https://docs.djangoproject.com/en/2.2/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

AUTH_USER_MODEL = "accounts.User"
# Password validation
# https://docs.djangoproject.com/en/2.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/2.2/topics/i18n/

LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.2/howto/static-files/

STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(BASE_DIR, 'data/static')

# LOGGING
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            # exact format is not important, this is the minimum information
            'format': '%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
        },
        'detailed': {
            'format': '[(%(levelname)s)] [%(asctime)s] %(message)s [%(name)s.%(funcName)s:%(lineno)s]'
        },
    },
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
        }
    },
    'handlers': {
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler'
        },
        'console': {
            'formatter': 'detailed',
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
        },
    },
    'loggers': {
        'django_test': {
            'handlers': ['console'],
            'level': 'DEBUG',
        },
        'django': {
            'handlers': ['console', 'mail_admins'],
            'level': 'INFO',
            'propagate': True,
        },
        'django.request': {
            'handlers': ['mail_admins'],
            'level': 'ERROR',
            'propagate': True,
        },
        # 'django.server': {
        # 'handlers': ['console'],
        # 'level': 'INFO',
        # 'propagate': False,
        # },
        'oidc_provider': {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': True,
        },
        # 'oidcop': {
        # 'handlers': ['console'],
        # 'level': 'DEBUG',
        # 'propagate': False,
        # },
        # 'oidcmsg': {
        # 'handlers': ['console'],
        # 'level': 'DEBUG',
        # 'propagate': False,
        # },
        # 'cryptojwt': {
        # 'handlers': ['console'],
        # 'level': 'DEBUG',
        # 'propagate': False,
        # },
    }
}



# Generated by Django 3.1.1 on 2021-05-11 10:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('oidc_provider', '0004_auto_20210511_0940'),
    ]

    operations = [
        migrations.AddField(
            model_name='oidcsession',
            name='sid_encrypted',
            field=models.CharField(blank=True, max_length=254, null=True),
        ),
    ]



import copy

from django.contrib.auth import authenticate
from django.contrib.auth import get_user_model
from django.template.loader import render_to_string

from oidcop.user_authn.user import (create_signed_jwt, LABELS)
from oidcop.user_authn.user import UserAuthnMethod

from . exceptions import FailedAuthentication


class UserPassDjango(UserAuthnMethod):
    """
    see oidcop.authn_context
        oidcop.endpoint_context
        https://docs.djangoproject.com/en/2.2/ref/templates/api/#rendering-a-context
    """

    # TODO: get this though settings conf
    url_endpoint = "/verify/user_pass_django"

    def __init__(self,
                 # template_handler=render_to_string,
                 template="oidc_login.html",
                 server_get=None, verify_endpoint='', **kwargs):
        """
        template_handler is only for backwards compatibility
        it will be always replaced by Django's default
        """
        super(UserPassDjango, self).__init__(server_get=server_get)

        self.kwargs = kwargs
        self.kwargs.setdefault("page_header", "Log in")
        self.kwargs.setdefault("user_label", "Username")
        self.kwargs.setdefault("passwd_label", "Password")
        self.kwargs.setdefault("submit_btn", "Log in")
        self.kwargs.setdefault("tos_uri", "")
        self.kwargs.setdefault("logo_uri", "")
        self.kwargs.setdefault("policy_uri", "")
        self.kwargs.setdefault("tos_label", "")
        self.kwargs.setdefault("logo_label", "")
        self.kwargs.setdefault("policy_label", "")

        # TODO this could be taken from args
        self.template_handler = render_to_string
        self.template = template

        self.action = verify_endpoint or self.url_endpoint
        self.kwargs['action'] = self.action

    def __call__(self, **kwargs):
        _ec = self.server_get('endpoint_context')
        # Stores information need afterwards in a signed JWT that then
        # appears as a hidden input in the form
        jws = create_signed_jwt(_ec.issuer, _ec.keyjar, **kwargs)

        self.kwargs['token'] = jws

        _kwargs = self.kwargs.copy()
        for attr in ['policy', 'tos', 'logo']:
            _uri = '{}_uri'.format(attr)
            _kwargs[_uri] = kwargs.get(_uri)
            _label = '{}_label'.format(attr)
            _kwargs[_label] = LABELS[_uri]

        return self.template_handler(self.template, _kwargs)

    def verify(self, *args, **kwargs):
        username = kwargs["username"]
        password = kwargs["password"]

        user = authenticate(username=username,
                            password=password)

        if username:
            return user
        else:
            raise FailedAuthentication()


class UserInfo(object):
    """ Read only interface to a user info store """

    def __init__(self, *args, **kwargs):
        self.claims_map = kwargs.get('claims_map', {})

    def filter(self, user, user_info_claims=None):
        """
        Return only those claims that are asked for.
        It's a best effort task; if essential claims are not present
        no error is flagged.

        :param userinfo: A dictionary containing available infos for one user
        :param user_info_claims: A dictionary specifying the asked for claims
        :return: A dictionary of filtered claims.
        """
        result = {}
        if not user.is_active:
            return result

        if user_info_claims is None:
            return copy.copy(user.__dict__)
        else:
            missing = []
            optional = []
            for key, restr in user_info_claims.items():
                if key in self.claims_map:
                    # manage required and optional: TODO extends this approach
                    if not hasattr(user, self.claims_map.get(key)) and \
                            restr == {"essential": True}:
                        missing.append(key)
                        continue
                    else:
                        optional.append(key)
                    #
                    uattr = getattr(user, self.claims_map[key], None)
                    if not uattr:
                        continue
                    result[key] = uattr() if callable(uattr) else uattr
            return result

    def __call__(self, user_id, client_id,
                 user_info_claims=None, **kwargs):
        """
        user_id = username
        client_id = client id, ex: 'mHwpZsDeWo5g'
        """
        user = get_user_model().objects.filter(username=user_id).first()
        if not user:  # pragma: no cover
            # Todo: raise exception here, this wouldn't be possible.
            return {}

        return self.filter(user, user_info_claims)



import logging
import os
import re
import requests
import subprocess
import urllib
import urllib3
import time

from cryptojwt import KeyJar
from cryptojwt.key_jar import init_key_jar

from oidcmsg.message import Message
from oidcrp import rp_handler
from oidcrp.util import load_yaml_config

from django.test import TestCase


logger = logging.getLogger('oidc_provider')
urllib3.disable_warnings()
RPHandler = rp_handler.RPHandler


def decode_token(bearer_token, keyjar, verify_sign=True):
    msg = Message().from_jwt(bearer_token,
                             keyjar=keyjar,
                             verify=verify_sign)
    return msg.to_dict()


def init_oidc_rp_handler(app):
    _rp_conf = app.config

    if _rp_conf.get('rp_keys'):
        _kj = init_key_jar(**_rp_conf['rp_keys'])
        _path = _rp_conf['rp_keys']['public_path']
        # removes ./ and / from the begin of the string
        _path = re.sub('^(.)/', '', _path)
    else:
        _kj = KeyJar()
        _path = ''
    _kj.httpc_params = _rp_conf['httpc_params']
    hash_seed = app.config.get('hash_seed', "BabyHoldOn")
    rph = RPHandler(_rp_conf['base_url'], _rp_conf['clients'], services=_rp_conf['services'],
                    hash_seed=hash_seed, keyjar=_kj, jwks_path=_path,
                    httpc_params=_rp_conf['httpc_params'])
    return rph


def get_rph(config_fname):
    config = load_yaml_config(config_fname)
    app = type('RPApplication', (object,), {"config": config})
    rph = init_oidc_rp_handler(app)
    return rph


def run_rp_session(rph, issuer_id, username, password):
    # register client (provider info and client registration)
    info = rph.begin(issuer_id)
    issuer_fqdn = rph.hash2issuer[issuer_id]
    issuer_keyjar = rph.issuer2rp[issuer_fqdn]

    print(f'Request authz grant: {info["url"]}\n')
    res = requests.get(info['url'], verify=rph.verify_ssl)

    auth_code = re.search(
        'value="(?P<token>[a-zA-Z\-\.\_0-9]*)"',
        res.text).groupdict()['token']
    auth_url = re.search(
        'action="(?P<url>[a-zA-Z0-9\/\-\_\.\:]*)"',
        res.text).groupdict()['url']

    auth_dict = {
        'username': username,
        'password': password,
        'token': auth_code
    }

    auth_url = ''.join((issuer_fqdn, auth_url))
    req = requests.post(auth_url,
                        data=auth_dict, verify=rph.verify_ssl,
                        allow_redirects=False)

    print("Submitting credentails form:")
    print(auth_url, auth_dict, '\n')

    # req is a 302, a redirect to the https://127.0.0.1:8099/authz_cb/django_oidc_op
    if req.status_code != 302:
        raise Exception(req.content)
    rp_final_url = req.headers['Location']
    ws, args = urllib.parse.splitquery(rp_final_url)

    request_args = urllib.parse.parse_qs(args)
    # from list to str
    request_args = {k: v[0] for k, v in request_args.items()}

    # oidcrp.RPHandler.finalize() will parse the authorization response and depending on the configuration run the needed RP service
    result = rph.finalize(request_args['iss'], request_args)

    # Tha't the access token, the bearer used to access to userinfo endpoint
    #  result['token']
    print("Bearer Access Token", result['token'], '\n')

    # get the keyjar related to the issuer
    decoded_access_token = decode_token(
        result['token'],
        keyjar=issuer_keyjar.get_service_context().keyjar
    )
    print("Access Token", decoded_access_token)
    print("ID Token", result['id_token'].to_dict())

    # userinfo
    result['userinfo'].to_dict()
    print("Userinfo endpoint result:", result['userinfo'].to_dict())


class TestOidcRPIntegration(TestCase):
    def setUp(self):
        self.rph = get_rph('../oidc_rp/conf.json')

        self.rph.verify_ssl = self.rph.httpc_params['verify']
        if not self.rph.verify_ssl:
            urllib3.disable_warnings()

        # select the OP you want to, example: "django_oidc_op"
        self.issuer_id = 'django_provider'

    def _run_example_project(self):
        """This is a pure integration test made with JWTConnect-Python-OidcRP"""
        os.chdir('../example')
        self.oidc_srv = subprocess.Popen(
            ["bash", "run.sh"],
            # stdout=subprocess.PIPE,
            # stderr=subprocess.PIPE
        )
        time.sleep(2)

    def test_all(self):
        self._run_example_project()
        run_rp_session(self.rph, self.issuer_id, 'test', 'testami18')
        os.system(
            """ps ax | grep "wsgi-file example/wsgi.py" | awk -F' ' {'print $1'} | xargs echo | xargs kill -KILL"""
        )



# Generated by Django 3.1.1 on 2021-05-11 09:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('oidc_provider', '0003_auto_20210510_1353'),
    ]

    operations = [
        migrations.AddField(
            model_name='oidcsession',
            name='key',
            field=models.CharField(blank=True, max_length=254, null=True),
        ),
        migrations.AlterField(
            model_name='oidcsession',
            name='salt',
            field=models.CharField(blank=True, max_length=254, null=True),
        ),
        migrations.AlterField(
            model_name='oidcsession',
            name='sid',
            field=models.CharField(blank=True, max_length=254, null=True),
        ),
        migrations.AlterField(
            model_name='oidcsession',
            name='sub',
            field=models.CharField(blank=True, max_length=254, null=True),
        ),
    ]



import re
import os
import sys

from glob import glob
from setuptools import setup, find_packages

SRC_FOLDER = '.'
PKG_NAME = 'oidc_provider'

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()


with open(os.path.join(os.path.dirname(__file__), 'requirements.txt')) as requirements:
    REQUIREMENTS = requirements.read()


setup(
    name="oidc_provider",
    version='2.0.1',
    description="django oidc provider",
    long_description=README,
    long_description_content_type='text/markdown',
    author='Giuseppe De Marco',
    author_email='demarcog83@gmail.com',
    license="Apache 2.0",
    url='https://github.com/peppelinux/django-oidc-op',
    packages=[PKG_NAME,],
    package_dir={PKG_NAME: f'{SRC_FOLDER}/{PKG_NAME}'},

    package_data={PKG_NAME: [i.replace(f'{SRC_FOLDER}/{PKG_NAME}/', '')
                             for i in glob(f'{SRC_FOLDER}/{PKG_NAME}/**',
                                           recursive=True)]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries :: Python Modules"],
    install_requires=REQUIREMENTS,
    zip_safe=False,
)



from django.urls import path

from . import views

app_name = 'oidc_provider'


urlpatterns = [
    path('.well-known/<str:service>', views.well_known,
         name="_well_known"),
    path('registration', views.registration,
         name="registration"),
    path('registration_read', views.registration_read,
         name="registration_read"),

    path('authorization', views.authorization,
         name="authorization"),

    path('verify/oidc_user_login/', views.verify_user,
         name="verify_user"),
    path('token', views.token, name="token"),
    path('userinfo', views.userinfo, name="userinfo"),
    path('introspection', views.introspection, name="introspection"),

    path('check_session_iframe', views.check_session_iframe,
         name="check_session_iframe"),
    path('session', views.session_endpoint, name="session"),
    # logout
    path('verify_logout', views.verify_logout,
         name="verify_logout"),
    path('post_logout', views.post_logout, name="post_logout"),
    path('rp_logout', views.rp_logout, name="rp_logout"),
]



DOMAIN = "127.0.0.1"
PORT = 8000
SERVER_NAME = f"{DOMAIN}:{PORT}"

OIDC_URL_PREFIX = 'oidcop/'

OIDC_SIGN_ALGS = [
    "RS256",
    "RS512",
    "ES256",
    "ES512",
    "PS256",
    "PS512",
]

OIDC_ENC_ALGS = [
    "RSA-OAEP",
    "RSA-OAEP-256",
    "A192KW",
    "A256KW",
    "ECDH-ES",
    "ECDH-ES+A128KW",
    "ECDH-ES+A192KW",
    "ECDH-ES+A256KW",
]

OIDC_ENCS = [
    'A128CBC-HS256',
    'A192CBC-HS384',
    'A256CBC-HS512',
    'A128GCM',
    'A192GCM',
    'A256GCM'
]

OIDC_KEY_DEFS = [
    {
        "type": "RSA",
        "use": [
            "sig"
        ]
    },
    {
        "type": "EC",
        "crv": "P-256",
        "use": [
            "sig"
        ]
    }
]

OIDC_OP_AUTHZ = {
    "class": "oidcop.authz.AuthzHandling",
    "kwargs": {
        "grant_config": {
            "usage_rules": {
                "authorization_code": {
                    "supports_minting": ["access_token", "refresh_token", "id_token"],
                    "max_usage": 1
                },
                "access_token": {},
                "refresh_token": {
                    "supports_minting": ["access_token", "refresh_token"]
                }
            },
            "expires_in": 43200
        }
    }
}

OIDC_JWKS_PRIVATE_PATH = 'data/oidc_op/private'
OIDC_JWKS_PUBLIC_PATH = 'data/static'


OIDC_OP_ENDPOINTS = {
    # not enabled :)
    # "webfinger": {
    # "path": ".well-known/webfinger",
    # "class": "oidcop.oidc.discovery.Discovery",
    # "kwargs": {
    # "client_authn_method": None
    # }
    # },
    "provider_info": {
        "path": ".well-known/openid-configuration",
        "class": "oidcop.oidc.provider_config.ProviderConfiguration",
        "kwargs": {
            "client_authn_method": None
        }
    },
    "registration": {
        "path": f"{OIDC_URL_PREFIX}registration",
        "class": "oidcop.oidc.registration.Registration",
        "kwargs": {
            "client_authn_method": None,
            "client_secret_expiration_time": 432000,
            "client_id_generator": {
                "class": 'oidcop.oidc.registration.random_client_id',
                "kwargs": {}
            }
        }
    },
    "registration_read": {
        "path": f"{OIDC_URL_PREFIX}registration_read",
        "class": "oidcop.oidc.read_registration.RegistrationRead",
        "kwargs": {
            "client_authn_method": [
                "bearer_header"
            ]
        }
    },
    "introspection": {
        "path": f"{OIDC_URL_PREFIX}introspection",
        "class": "oidcop.oauth2.introspection.Introspection",
        "kwargs": {
            "client_authn_method": [
                "client_secret_post",
                "client_secret_basic",
                "client_secret_jwt",
                "private_key_jwt"
            ],
            "release": [
                "username"
            ]
        }
    },
    "authorization": {
        "path": f"{OIDC_URL_PREFIX}authorization",
        "class": "oidcop.oidc.authorization.Authorization",
        "kwargs": {
            "client_authn_method": None,
            "claims_parameter_supported": True,
            "request_parameter_supported": True,
            "request_uri_parameter_supported": True,

            "request_object_encryption_alg_values_supported": OIDC_ENC_ALGS,

            "response_types_supported": [
                "code",
                # "token",
                # "id_token",
                # "code token",
                # "code id_token",
                # "id_token token",
                # "code id_token token",
                # "none"
            ],
            "response_modes_supported": [
                "query",
                "fragment",
                "form_post"
            ]
        }
    },
    "token": {
        "path": f"{OIDC_URL_PREFIX}token",
        "class": "oidcop.oidc.token.Token",
        "kwargs": {
            "client_authn_method": [
                "client_secret_post",
                "client_secret_basic",
                "client_secret_jwt",
                "private_key_jwt"
            ],
        }
    },
    "userinfo": {
        "path": f"{OIDC_URL_PREFIX}userinfo",
        "class": "oidcop.oidc.userinfo.UserInfo",
        "kwargs": {
            "claim_types_supported": [
                "normal",
                "aggregated",
                "distributed"
            ],
            "userinfo_signing_alg_values_supported": OIDC_SIGN_ALGS,
            "userinfo_encryption_alg_values_supported": OIDC_ENC_ALGS,
        }
    },
    "end_session": {
        "path": f"{OIDC_URL_PREFIX}session",
        "class": "oidcop.oidc.session.Session",
        "kwargs": {
            "logout_verify_url": "verify_logout",
            "post_logout_uri_path": "post_logout",
            "signing_alg": "ES256",
            "frontchannel_logout_supported": True,
            "frontchannel_logout_session_supported": True,
            "backchannel_logout_supported": True,
            "backchannel_logout_session_supported": True,
            # "check_session_iframe": "check_session_iframe"
        }
    }
}


OIDC_OP_TOKEN_HANDLER = {
    "jwks_file": f"{OIDC_JWKS_PRIVATE_PATH}/token_jwks.json",
    "code": {
        "kwargs": {
            "lifetime": 600
        }
    },
    "token": {
        "class": "oidcop.token.jwt_token.JWTToken",
        "kwargs": {
            "lifetime": 3600,
            # this will add email in access_token
            # "base_claims": {
            # "email": None,
            # "email_verified": None,
            # },
            # "enable_claims_per_client": True,
            # "aud": [
            # "https://example.org/appl"
            # ]
        }
    },
    "refresh": {
        "kwargs": {
            "lifetime": 86400
        }
    },
    "id_token": {
        "class": "oidcop.token.id_token.IDToken",
        "kwargs": {
            "id_token_signing_alg_values_supported": OIDC_SIGN_ALGS,
            "id_token_encryption_alg_values_supported": OIDC_ENC_ALGS,
            "id_token_encryption_enc_values_supported": OIDC_ENCS,
        }
    }
}

OIDCOP_CONFIG = {
    "port": PORT,
    "domain": DOMAIN,
    "server_name": SERVER_NAME,
    "base_url": f"https://{SERVER_NAME}",
    "op": {
        "server_info": {
            "add_on": {
                "pkce": {
                    "function": "oidcop.oidc.add_on.pkce.add_pkce_support",
                    "kwargs": {
                        "essential": False,
                        "code_challenge_method": "S256 S384 S512"
                    }
                },
                "claims": {
                    "function": "oidcop.oidc.add_on.custom_scopes.add_custom_scopes",
                    "kwargs": {
                        "research_and_scholarship": [
                            "name",
                            "given_name",
                            "family_name",
                            "email",
                            "email_verified",
                            "sub",
                            "iss",
                            "eduperson_scoped_affiliation"
                        ],
                        # "profile": [
                        # "email"
                        # ]
                    }
                }
            },
            "authz": OIDC_OP_AUTHZ,
            "authentication": {
                "user": {
                    "acr": "urn:oasis:names:tc:SAML:2.0:ac:classes:InternetProtocolPassword",
                    "class": "oidc_provider.users.UserPassDjango",
                    "kwargs": {
                        "verify_endpoint": f"{OIDC_URL_PREFIX}verify/oidc_user_login/",
                        "template": "oidc_login.html",

                        "page_header": "Testing log in",
                        "submit_btn": "Get me in!",
                        "user_label": "Nickname",
                        "passwd_label": "Secret sauce"
                    }
                }
            },
            "capabilities": {
                "subject_types_supported": [
                    "public",
                    "pairwise"
                ],
                "grant_types_supported": [
                    "authorization_code",
                    # "implicit",
                    "urn:ietf:params:oauth:grant-type:jwt-bearer",
                    "refresh_token"
                ],
                # indicates that unknow/unavailable scopes requested by a RP
                # would get a 403 error message instead of be declined implicitly.
                # If False the op will only release the available scopes and ignoring the missings.
                # Default to False
                # deny_unknown_scopes: True
            },
            "cookie_handler": {
                "class": "oidcop.cookie_handler.CookieHandler",
                "kwargs": {
                    "keys": {
                        "private_path": f"{OIDC_JWKS_PRIVATE_PATH}/cookie_jwks.json",
                        "key_defs": [
                          {"type": "OCT", "use": ["enc"], "kid": "enc"},
                          {"type": "OCT", "use": ["sig"], "kid": "sig"}
                        ],
                        "read_only": False
                    },
                    "flags": {
                        "samesite": "None",
                        "httponly": True,
                        "secure": True,
                    },
                    "name": {
                        "session": "oidc_op",
                        "register": "oidc_op_rp",
                        "session_management": "sman"
                    }
                }
            },
            "httpc_params": {
                "verify": False
            },
            "endpoint": OIDC_OP_ENDPOINTS,
            "issuer": f"https://{SERVER_NAME}",
            "keys": {
                "private_path": f"{OIDC_JWKS_PRIVATE_PATH}/jwks.json",
                "key_defs": OIDC_KEY_DEFS,
                "public_path": f"{OIDC_JWKS_PUBLIC_PATH}/jwks.json",
                "read_only": False,
                "uri_path": "static/jwks.json"
            },
            "login_hint2acrs": {
                "class": "oidcop.login_hint.LoginHint2Acrs",
                "kwargs": {
                    "scheme_map": {
                        "email": [
                            "urn:oasis:names:tc:SAML:2.0:ac:classes:InternetProtocolPassword"
                        ]
                    }
                }
            },
            "session_params": {
                "password": '__password_used_to_encrypt_access_token_sid_value',
                "salt": "salt involved in session sub hash ",
                "sub_func": {
                    "public": {
                        "class": "oidcop.session.manager.PublicID",
                        "kwargs": {
                            "salt": "sdfsdfdsf"
                        }
                    },
                    "pairwise": {
                        "class": "oidcop.session.manager.PairWiseID",
                        "kwargs": {
                            "salt": "sdfsdfsdf"
                        }
                    }
                },
            },
            "template_dir": "templates",
            "token_handler_args": OIDC_OP_TOKEN_HANDLER,
            "userinfo": {
                "class": "oidc_provider.users.UserInfo",
                "kwargs": {
                    # map claims to django user attributes here:
                    "claims_map": {
                        "phone_number": "telephone",
                        "family_name": "last_name",
                        "given_name": "first_name",
                        "email": "email",
                        "verified_email": "email",
                        "gender": "gender",
                        "birthdate": "get_oidc_birthdate",
                        "updated_at": "get_oidc_lastlogin"
                    }
                }
            }
        }
    }

}



# Generated by Django 3.0 on 2021-05-22 13:22

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('oidc_provider', '0009_auto_20210517_1803'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='oidcsession',
            name='session_info',
        ),
    ]
